###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""CacheExplorer implementation."""
from __future__ import annotations

__all__ = (
    'CacheIndexContextMenu',
    'file_to_clipboard',
)

from pathlib import Path

from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *

from ...constants import *
from ...models import DeferredCallable
from ...utils import add_menu_items
from ...utils import init_objects
from ..aliases import app
from ..aliases import tr
from ..widgets import CacheExplorer


def file_to_clipboard(file_path: Path):
    """Copy the file's contents to the application's clipboard.

    Works with both images and text.

    :param file_path: File path to read from.
    """
    if str(file_path).endswith(tuple(SUPPORTED_IMAGE_EXTENSIONS)):
        pixmap = QPixmap()
        pixmap.loadFromData(file_path.read_bytes())
        app().clipboard().setImage(pixmap.toImage())
    else:
        app().clipboard().setText(file_path.read_text(encoding='utf8'))


# noinspection PyProtectedMember
class CacheIndexContextMenu(QMenu):
    """Context menu that shows actions to manipulate files."""

    def __init__(self, parent: CacheExplorer, index: QModelIndex):
        """Create a new :py:class:`CacheIndexContextMenu` for the given parent and index."""
        super().__init__(parent)

        file_path = Path(parent.model().filePath(index))
        if not (dir_path := file_path).is_dir():
            dir_path = file_path.parent

        try:
            os_path = app().client.to_get_path(file_path.as_posix())
        except (IndexError, ValueError):
            os_path = ''

        init_objects({
            (folding_submenu := QMenu(self)): {
                'title': tr('gui.menus.cached_file.folding')
            },

            (open_in_view := QAction(self)): {
                'disabled': file_path.is_dir(),
                'text': tr('gui.menus.cached_file.open_in_view'),
                'icon': self.style().standardIcon(QStyle.StandardPixmap.SP_DesktopIcon),
                'triggered': DeferredCallable(
                    parent.openFileInView.emit, file_path.as_posix()
                )
            },

            (open_in_default_app := QAction(self)): {
                'disabled': file_path.is_dir(),
                'text': tr('gui.menus.cached_file.open_in_default_app'),
                'icon': parent.model().iconProvider()._fallback_provider.icon(  # type: ignore
                    parent.model().fileInfo(index)
                ),
                'triggered': DeferredCallable(
                    QDesktopServices.openUrl,
                    QUrl(file_path.as_uri())
                )
            },

            (open_in_explorer := QAction(self)): {
                'text': tr('gui.menus.cached_file.view_in_explorer'),
                'icon': (app().get_theme_icon('dialog_open') or
                         app().icon_store['folder']),
                'triggered': DeferredCallable(
                    QDesktopServices.openUrl,
                    QUrl(dir_path.as_uri())
                )
            },

            (expand_this := QAction(self)): {
                'disabled': file_path.is_file(),
                'text': tr('gui.menus.cached_file.folding.expand_this'),
                'triggered': DeferredCallable(parent.expand, index)
            },

            (expand_recursively := QAction(self)): {
                'disabled': file_path.is_file(),
                'text': tr('gui.menus.cached_file.folding.expand_recursively'),
                'triggered': DeferredCallable(parent.expandRecursively, index)
            },

            (expand_all := QAction(self)): {
                'text': tr('gui.menus.cached_file.folding.expand_all'),
                'triggered': parent.expandAll
            },

            (collapse_this := QAction(self)): {
                'disabled': file_path.is_file(),
                'text': tr('gui.menus.cached_file.folding.collapse_this'),
                'triggered': DeferredCallable(parent.collapse, index)
            },

            (collapse_all := QAction(self)): {
                'text': tr('gui.menus.cached_file.folding.collapse_all'),
                'triggered': parent.collapseAll
            },

            (copy_full_path := QAction(self)): {
                'text': tr('gui.menus.cached_file.copy_full_path'),
                'triggered': DeferredCallable(app().clipboard().setText, str(file_path))
            },

            (copy_endpoint_path := QAction(self)): {
                'disabled': not os_path,
                'text': tr('gui.menus.cached_file.copy_endpoint_path'),
                'triggered': DeferredCallable(app().clipboard().setText, os_path)
            },

            (copy_contents := QAction(self)): {
                'disabled': file_path.is_dir(),
                'text': tr('gui.menus.cached_file.copy_contents'),
                'triggered': DeferredCallable(file_to_clipboard, file_path)
            },

            (delete := QAction(self)): {
                'text': tr('gui.menus.cached_file.delete', 'Folder' if file_path.is_dir() else 'File'),
                'icon': (app().get_theme_icon('dialog_cancel') or
                         self.style().standardIcon(QStyle.StandardPixmap.SP_DialogCancelButton)),
                'triggered': DeferredCallable(parent.delete_index, index)
            },
        })

        add_menu_items(folding_submenu, [
            'Expand', expand_this, expand_recursively, expand_all,
            'Collapse', collapse_this, collapse_all,
        ])

        add_menu_items(self, [
            'Open', open_in_view, open_in_default_app, open_in_explorer,
            'Folding', folding_submenu,
            'Copy', copy_full_path, copy_endpoint_path, copy_contents,
            'Delete', delete
        ])



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""Package for managing outgoing requests for hi_getter."""
from __future__ import annotations

__all__ = (
    'gc_response',
    'KNOWN_HEADERS',
    'NetworkSession',
    'Request',
    'Response',
)

import datetime as dt
import json as json_
import re
from collections.abc import Callable
from collections.abc import Mapping
from collections.abc import Sequence
from json import dumps as json_dumps
from pathlib import Path
from typing import Any
from typing import Final
from typing import TypeAlias
from warnings import warn
from weakref import WeakKeyDictionary

from PySide6.QtCore import *
from PySide6.QtNetwork import *
from shiboken6 import Shiboken

from ..models import CaseInsensitiveDict
from ..models import DeferredCallable
from ..utils import dict_to_query
from ..utils import encode_url_params
from ..utils import guess_json_utf
from ..utils import is_error_status
from ..utils import query_to_dict
from ..utils import wait_for_reply

_StringPair: TypeAlias = dict[str, str] | list[tuple[str, str]]
_KnownHeaderValues: TypeAlias = str | bytes | dt.datetime | dt.date | dt.time | _StringPair | list[str]
_HeaderValue: TypeAlias = dict[str, _KnownHeaderValues] | list[tuple[str, _KnownHeaderValues]]

_INT_PATTERN: Final[re.Pattern] = re.compile(r'[1-9]\d*|0')


# autopep8: off
KNOWN_HEADERS: CaseInsensitiveDict[tuple[QNetworkRequest.KnownHeaders, type]] = CaseInsensitiveDict({
    # Name:                Header Enum Value:                                      Object Wanted:
    # -----------------------------------------------------------------------------------------------
    'Content-Disposition': (QNetworkRequest.KnownHeaders.ContentDispositionHeader, str),
    'Content-Type':        (QNetworkRequest.KnownHeaders.ContentTypeHeader,        str),
    'Content-Length':      (QNetworkRequest.KnownHeaders.ContentLengthHeader,      bytes),
    'Cookie':              (QNetworkRequest.KnownHeaders.CookieHeader,             QNetworkCookie),
    'ETag':                (QNetworkRequest.KnownHeaders.ETagHeader,               str),
    'If-Match':            (QNetworkRequest.KnownHeaders.IfMatchHeader,            QStringListModel),
    'If-Modified-Since':   (QNetworkRequest.KnownHeaders.IfModifiedSinceHeader,    QDateTime),
    'If-None-Match':       (QNetworkRequest.KnownHeaders.IfNoneMatchHeader,        QStringListModel),
    'Last-Modified':       (QNetworkRequest.KnownHeaders.LastModifiedHeader,       QDateTime),
    'Location':            (QNetworkRequest.KnownHeaders.LocationHeader,           QUrl),
    'Server':              (QNetworkRequest.KnownHeaders.ServerHeader,             str),
    'Set-Cookie':          (QNetworkRequest.KnownHeaders.SetCookieHeader,          QNetworkCookie),
    'User-Agent':          (QNetworkRequest.KnownHeaders.UserAgentHeader,          str),
})
# autopep8: on


def _translate_header_value(
        header: str, value: _KnownHeaderValues
) -> _KnownHeaderValues | QDateTime | list[QNetworkCookie] | QUrl:
    """Translate a header's value to it's appropriate type for use in QNetworkRequest.setHeader.

    Values are translated to their appropriate type based on the
    type defined in KNOWN_HEADERS next to the header enum value.

    The following types are supported:
        - str: Given value is translated into a str.
        - bytes: Translates string value into a utf8 encoded version.
        - QDateTime: Translates string and datetime values into a QDateTime.
        - QNetworkCookie: Translates string pairs into a QNetworkCookie list.
          The first value is the cookie name, the second is the cookie value.
        - QStringListModel: Iterates over value and translates all inner-values to strings.
          Returns a list of the translated strings.
        - QUrl: Calls the QUrl constructor on value and returns result.

    :param header: Header defined in KNOWN_HEADERS.
    :param value: Value to translate into an accepted type.
    :return: Transformed value.
    """
    # Match the known-header's value name and translate value to that type.
    match KNOWN_HEADERS[header][1].__name__:
        case 'str':
            return str(value)

        case 'bytes':
            if isinstance(value, str):
                return value.encode('utf8')

        case 'QDateTime':
            if isinstance(value, (dt.datetime, dt.date, dt.time)):
                date_value: dt.datetime | dt.date | dt.time = value

                # Translate datetime objects to a string
                if not isinstance(value, dt.datetime):
                    date: dt.date = dt.datetime.now().date() if isinstance(value, dt.time) else value
                    time: dt.time = dt.datetime.now().time() if isinstance(value, dt.date) else value
                    date_value = dt.datetime.fromisoformat(f'{date.isoformat()}T{time.isoformat()}')

                return QDateTime().fromString(date_value.isoformat(), Qt.DateFormat.ISODateWithMs)

            # Translate string to QDateTime object
            return QDateTime().fromString(str(value), Qt.DateFormat.ISODateWithMs)

        case 'QNetworkCookie':
            cookie_list: list[QNetworkCookie] = []
            # Translate mappings
            if isinstance(value, Mapping):
                for name, _value in value.items():
                    cookie_list.append(QNetworkCookie(name.encode('utf8'), _value.encode('utf8')))

            # Translate tuples, lists, etc. that contain two strings (name and value)
            elif isinstance(value, Sequence) and not isinstance(value, (bytes, str)):
                for pair in value:
                    cookie_list.append(QNetworkCookie(pair[0].encode('utf8'), pair[1].encode('utf8')))

            return cookie_list

        case 'QStringListModel':
            if isinstance(value, Sequence):
                return [str(item) for item in value]

        case 'QUrl':
            if not isinstance(value, QUrl):
                return QUrl(str(value))

    return value


def gc_response(func: Callable[[Response], Any]) -> Callable[[Response], Any]:
    """Wrap the given function to delete a :py:class:`Response` after being called.

    This should only be used for one-time calls, such as a "handling reply" function.
    """
    def wrapper(response: Response, *args, **kwargs) -> Any:
        """Call ``delete()`` on first argument after being called."""
        ret_val: Any = func(response, *args, **kwargs)
        response.delete()
        return ret_val

    return wrapper


class NetworkSession:
    """``requests``-like wrapper over a :py:class:`QNetworkAccessManager`.

    The following convenience methods are supported:
        - get
        - head
        - post
        - put
        - delete
        - patch
    """

    def __init__(self, manager_parent: QObject | None = None) -> None:
        """Initialize the NetworkSession.

        :param manager_parent: Parent of the QNetworkAccessManager.
        """
        self._headers = CaseInsensitiveDict()
        self.manager = QNetworkAccessManager(manager_parent)
        self.default_redirect_policy = QNetworkRequest.RedirectPolicy.UserVerifiedRedirectPolicy
        self.reply_auth_map: WeakKeyDictionary[QNetworkReply, tuple[str, str]] = WeakKeyDictionary()

        self.manager.authenticationRequired.connect(self._handle_auth)  # pyright: ignore[reportGeneralTypeIssues]

    @property
    def cookies(self) -> dict[str, str]:
        """Return dictionary representation of the internal QNetworkCookieJar."""
        return {cookie.name().toStdString(): cookie.value().toStdString() for
                cookie in self.manager.cookieJar().allCookies()}

    @cookies.deleter
    def cookies(self) -> None:
        """Clear all cookies on delete."""
        self.clear_cookies()

    @property
    def headers(self) -> CaseInsensitiveDict[Any]:
        """:return: Dictionary containing the default session headers."""
        return self._headers

    @headers.setter
    def headers(self, value: Mapping) -> None:
        """Translate any mapping value to a CaseInsensitiveDict for use as headers.

        :param value: Mapping to copy into a CaseInsensitiveDict.
        :raises TypeError: If the value is not a Mapping.
        """
        if not isinstance(value, Mapping):
            raise TypeError(f'NetworkSession headers must be a Mapping, not {type(value)}')

        self._headers = CaseInsensitiveDict(value)

    @headers.deleter
    def headers(self) -> None:
        """Clear headers on delete."""
        self._headers.clear()

    @staticmethod
    def _check_method_kwargs(method: str, **kwargs) -> None:
        """Check that the given keyword arguments are valid for the given HTTP method.

        If some arguments are invalid, a warning is emitted.

        :param method: HTTP method to check.
        :param kwargs: Keyword arguments to check.
        """
        if method in {'GET', 'HEAD', 'CONNECT', 'OPTIONS', 'TRACE'}:
            if any((kwargs.get('data'), kwargs.get('files'), kwargs.get('json'))):
                warn(UserWarning(
                    f'{method} requests do not support data attached to the request body. '
                    f'This data is likely to be ignored.'
                ))

    def _handle_auth(self, reply: QNetworkReply, authenticator: QAuthenticator) -> None:
        if reply in self.reply_auth_map:
            user, password = self.reply_auth_map[reply]
            authenticator.setUser(user)
            authenticator.setPassword(password)

    def clear_cookies(self, domain: str | None = None, path: str | None = None, name: str | None = None) -> bool:
        """Clear some cookies. Functionally equivalent to http.cookiejar.clear.

        Invoking this method without arguments will clear all cookies.  If
        given a single argument, only cookies belonging to that domain will be
        removed.  If given two arguments, cookies belonging to the specified
        path within that domain are removed.  If given three arguments, then
        the cookie with the specified name, path and domain is removed.

        :param domain: The domain of the cookie to remove.
        :param path: The path of the cookie to remove.
        :param name: The name of the cookie to remove.
        :return: True if any cookies were removed, False otherwise.
        :raises ValueError: If name is provided, must provide path.
          If path is provided, must provide domain. Raise otherwise.
        """

        def deletion_predicate(cookie: QNetworkCookie) -> bool:
            """Return whether the cookie should be deleted.

            :param cookie: The cookie to check.
            :return: True if the cookie should be removed, False otherwise.
            :raises ValueError: If name is provided, must provide path.
              If path is provided, must provide domain. Raise otherwise.
            """
            # 3 args -- Delete the specific cookie which matches all information.
            if name is not None:
                if domain is None or path is None:
                    raise ValueError('Must specify domain and path if specifying name')

                return cookie.name().toStdString() == name and cookie.domain() == domain and cookie.path() == path

            # 2 args -- Delete all cookies with the given domain and path.
            if path is not None:
                if domain is None:
                    raise ValueError('Must specify domain if specifying path')

                return cookie.domain() == domain and cookie.path() == path

            # 1 arg -- Delete all cookies in the given domain.
            if domain is not None:
                return cookie.domain() == domain

            # 0 args -- Delete all cookies
            return True

        results = []
        for _cookie in self.manager.cookieJar().allCookies():
            if deletion_predicate(_cookie):
                results.append(self.manager.cookieJar().deleteCookie(_cookie))

        return any(results)

    def set_cookie(self, name: str, value: str, domain: str, path: str | None = None) -> bool:
        """Create a new cookie with the given date.

        Replaces a pre-existing cookie with the same identifier if it exists.

        :return: True if the cookie was set.
        :raises: ValueError if name and value are not strings.
        """
        cookie = QNetworkCookie(name=name.encode('utf8'), value=value.encode('utf8'))
        cookie.setDomain(domain)
        cookie.setPath(path or '/')
        return self.manager.cookieJar().insertCookie(cookie)

    def request(self, method: str, url: QUrl | str, *args, **kwargs) -> Response:
        """Send an HTTP request to the given URL with the given data.

        -----

        See :py:meth:`Request.__init__` for full kwarg documentation.

        :param method: HTTP method/verb to use for the request. Case-sensitive.
        :param url: URL to send the request to. Case-sensitive.
        :keyword params: URL parameters to attach to the URL. Case-sensitive.
        :keyword data: Bytes to send in the request body.
        :keyword headers: Headers to use for the request. Case-insensitive.
        :keyword cookies: Cookies to use for the request. Case-sensitive.
        :keyword auth: Optional tuple containing username and password.
        :keyword timeout: Timeouts for the request.
        :keyword allow_redirects: If False, do not follow any redirect requests.
        :keyword proxies: String-pairs mapping protocol to the URL of the proxy.
        :keyword stream: Whether to accept chunked encoding.
        :keyword verify: Whether to verify SSL certificates.
        :keyword cert: Client certificate information.
        :keyword json: JSON data to send in the request body.
        :keyword wait_until_finished: Process the application eventLoop until the reply is finished.
        :keyword finished: Callback when the request finishes, with request supplied as an argument.
        :keyword progress: Callback to update download progress, with the request, received bytes, and total bytes.

        :return: Response object, which is not guaranteed to be finished.
        """
        send_kwargs = {
            'wait_until_finished': kwargs.pop('wait_until_finished', False),
            'finished': kwargs.pop('finished', None),
            'progress': kwargs.pop('progress', None),
        }

        return Request(method, url, *args, **kwargs).send(self, **send_kwargs)

    def get(self, url: QUrl | str, **kwargs) -> Response:
        """Create and send a request with the GET HTTP method.

        GET is the general method used to get a resource from a server.
        It is the most commonly used method, with GET requests being used by web browsers to
        download HTML pages, images, and other resources.

        -----

        See :py:meth:`Request.__init__` for full kwarg documentation.

        :param url: URL to send the request to. Case-sensitive.
        :keyword params: URL parameters to attach to the URL. Case-sensitive.
        :keyword data: Bytes to send in the request body.
        :keyword headers: Headers to use for the request. Case-insensitive.
        :keyword cookies: Cookies to use for the request. Case-sensitive.
        :keyword auth: Optional tuple containing username and password.
        :keyword timeout: Timeouts for the request.
        :keyword allow_redirects: If False, do not follow any redirect requests.
        :keyword proxies: String-pairs mapping protocol to the URL of the proxy.
        :keyword stream: Whether to accept chunked encoding.
        :keyword verify: Whether to verify SSL certificates.
        :keyword cert: Client certificate information.
        :keyword json: JSON data to send in the request body.
        :keyword wait_until_finished: Process the application eventLoop until the reply is finished.
        :keyword finished: Callback when the request finishes, with request supplied as an argument.
        :keyword progress: Callback to update download progress, with the request, received bytes, and total bytes.

        :return: Response object, which is not guaranteed to be finished.
        """
        method: str = 'GET'
        self._check_method_kwargs(method, **kwargs)

        return self.request(method=method, url=url, **kwargs)

    def head(self, url: QUrl | str, **kwargs) -> Response:
        """Create and send a request with the HEAD HTTP method.

        HEAD requests are used to retrieve information about a resource without actually fetching the resource itself.
        This is useful for checking if a resource exists, or for getting the size of a resource before downloading it.

        -----

        See :py:meth:`Request.__init__` for full kwarg documentation.

        :param url: URL to send the request to. Case-sensitive.
        :keyword params: URL parameters to attach to the URL. Case-sensitive.
        :keyword data: Bytes to send in the request body.
        :keyword headers: Headers to use for the request. Case-insensitive.
        :keyword cookies: Cookies to use for the request. Case-sensitive.
        :keyword auth: Optional tuple containing username and password.
        :keyword timeout: Timeouts for the request.
        :keyword allow_redirects: If False, do not follow any redirect requests.
        :keyword proxies: String-pairs mapping protocol to the URL of the proxy.
        :keyword stream: Whether to accept chunked encoding.
        :keyword verify: Whether to verify SSL certificates.
        :keyword cert: Client certificate information.
        :keyword json: JSON data to send in the request body.
        :keyword wait_until_finished: Process the application eventLoop until the reply is finished.
        :keyword finished: Callback when the request finishes, with request supplied as an argument.
        :keyword progress: Callback to update download progress, with the request, received bytes, and total bytes.

        :return: Response object, which is not guaranteed to be finished.
        """
        method: str = 'HEAD'
        self._check_method_kwargs(method, **kwargs)

        return self.request(method=method, url=url, **kwargs)

    def post(self, url: QUrl | str, **kwargs) -> Response:
        """Create and send a request with the POST HTTP method.

        POST is the general method used to send data to a server.
        It does not require a resource to previously exist, nor does it require one to not exist.
        This makes it very common for servers to accept POST requests for a multitude of things.

        -----

        See :py:meth:`Request.__init__` for full kwarg documentation.

        :param url: URL to send the request to. Case-sensitive.
        :keyword params: URL parameters to attach to the URL. Case-sensitive.
        :keyword data: Bytes to send in the request body.
        :keyword headers: Headers to use for the request. Case-insensitive.
        :keyword cookies: Cookies to use for the request. Case-sensitive.
        :keyword auth: Optional tuple containing username and password.
        :keyword timeout: Timeouts for the request.
        :keyword allow_redirects: If False, do not follow any redirect requests.
        :keyword proxies: String-pairs mapping protocol to the URL of the proxy.
        :keyword stream: Whether to accept chunked encoding.
        :keyword verify: Whether to verify SSL certificates.
        :keyword cert: Client certificate information.
        :keyword json: JSON data to send in the request body.
        :keyword wait_until_finished: Process the application eventLoop until the reply is finished.
        :keyword finished: Callback when the request finishes, with request supplied as an argument.
        :keyword progress: Callback to update download progress, with the request, received bytes, and total bytes.

        :return: Response object, which is not guaranteed to be finished.
        """
        method: str = 'POST'
        self._check_method_kwargs(method, **kwargs)

        return self.request(method=method, url=url, **kwargs)

    def put(self, url: QUrl | str, **kwargs) -> Response:
        """Create and send a request with the PUT HTTP method.

        PUT is a method for completely updating a resource on a server.
        The data sent by PUT should be the full content of the resource.

        -----

        See :py:meth:`Request.__init__` for full kwarg documentation.

        :param url: URL to send the request to. Case-sensitive.
        :keyword params: URL parameters to attach to the URL. Case-sensitive.
        :keyword data: Bytes to send in the request body.
        :keyword headers: Headers to use for the request. Case-insensitive.
        :keyword cookies: Cookies to use for the request. Case-sensitive.
        :keyword auth: Optional tuple containing username and password.
        :keyword timeout: Timeouts for the request.
        :keyword allow_redirects: If False, do not follow any redirect requests.
        :keyword proxies: String-pairs mapping protocol to the URL of the proxy.
        :keyword stream: Whether to accept chunked encoding.
        :keyword verify: Whether to verify SSL certificates.
        :keyword cert: Client certificate information.
        :keyword json: JSON data to send in the request body.
        :keyword wait_until_finished: Process the application eventLoop until the reply is finished.
        :keyword finished: Callback when the request finishes, with request supplied as an argument.
        :keyword progress: Callback to update download progress, with the request, received bytes, and total bytes.

        :return: Response object, which is not guaranteed to be finished.
        """
        method: str = 'PUT'
        self._check_method_kwargs(method, **kwargs)

        return self.request(method=method, url=url, **kwargs)

    def delete(self, url: QUrl | str, **kwargs) -> Response:
        """Create and send a request with the DELETE HTTP method.

        DELETE is used to delete a specified resource.

        -----

        See :py:meth:`Request.__init__` for full kwarg documentation.

        :param url: URL to send the request to. Case-sensitive.
        :keyword params: URL parameters to attach to the URL. Case-sensitive.
        :keyword data: Bytes to send in the request body.
        :keyword headers: Headers to use for the request. Case-insensitive.
        :keyword cookies: Cookies to use for the request. Case-sensitive.
        :keyword auth: Optional tuple containing username and password.
        :keyword timeout: Timeouts for the request.
        :keyword allow_redirects: If False, do not follow any redirect requests.
        :keyword proxies: String-pairs mapping protocol to the URL of the proxy.
        :keyword stream: Whether to accept chunked encoding.
        :keyword verify: Whether to verify SSL certificates.
        :keyword cert: Client certificate information.
        :keyword json: JSON data to send in the request body.
        :keyword wait_until_finished: Process the application eventLoop until the reply is finished.
        :keyword finished: Callback when the request finishes, with request supplied as an argument.
        :keyword progress: Callback to update download progress, with the request, received bytes, and total bytes.

        :return: Response object, which is not guaranteed to be finished.
        """
        method: str = 'DELETE'
        self._check_method_kwargs(method, **kwargs)

        return self.request(method=method, url=url, **kwargs)

    def patch(self, url: QUrl | str, **kwargs) -> Response:
        """Create and send a request with the PATCH HTTP method.

        PATCH is used to send a partial update of an existing resource.

        -----

        See :py:meth:`Request.__init__` for full kwarg documentation.

        :param url: URL to send the request to. Case-sensitive.
        :keyword params: URL parameters to attach to the URL. Case-sensitive.
        :keyword data: Bytes to send in the request body.
        :keyword headers: Headers to use for the request. Case-insensitive.
        :keyword cookies: Cookies to use for the request. Case-sensitive.
        :keyword auth: Optional tuple containing username and password.
        :keyword timeout: Timeouts for the request.
        :keyword allow_redirects: If False, do not follow any redirect requests.
        :keyword proxies: String-pairs mapping protocol to the URL of the proxy.
        :keyword stream: Whether to accept chunked encoding.
        :keyword verify: Whether to verify SSL certificates.
        :keyword cert: Client certificate information.
        :keyword json: JSON data to send in the request body.
        :keyword wait_until_finished: Process the application eventLoop until the reply is finished.
        :keyword finished: Callback when the request finishes, with request supplied as an argument.
        :keyword progress: Callback to update download progress, with the request, received bytes, and total bytes.

        :return: Response object, which is not guaranteed to be finished.
        """
        method: str = 'PATCH'
        self._check_method_kwargs(method, **kwargs)

        return self.request(method=method, url=url, **kwargs)


class Request:
    """``requests``-like wrapper over a :py:class:`QNetworkRequest`."""

    def __init__(self, method: str, url: QUrl | str,
                 params: _StringPair | None = None,
                 data: bytes | _StringPair | None = None,
                 headers: _HeaderValue | None = None,
                 cookies: _StringPair | None = None,
                 # TODO: Finish ``requests``-like implementation
                 # files: dict[str, Any] | None = None,
                 auth: tuple[str, str] | None = None,
                 timeout: float | tuple[float, float] | None = 30.0,
                 allow_redirects: bool = True,
                 proxies: _StringPair | None = None,
                 stream: bool = False,
                 verify: bool | str | None = None,
                 cert: str | tuple[str, str] | None = None,
                 json: dict[str, Any] | None = None) -> None:
        """Create a new :py:class:`Request` with the given fields.

        :param method: HTTP method/verb to use for the request. Case-sensitive.

        :param url: URL to send the request to. Case-sensitive.
            Could be a string or QUrl.

        :param params: URL parameters to attach to the URL. Case-sensitive.
            If url is a QUrl, overrides the QUrl's query.

        :param data: Bytes to send in the request body.
            If a string-pair, will be encoded to bytes as a form-encoded request body.
            Incompatible with the json and files parameters.

        :param headers: Headers to use for the request.
            Non-string values should ONLY be used for KNOWN_HEADERS. Case-insensitive.

        :param cookies: Cookies to use for the request. Case-sensitive.

        :param auth: Optional tuple containing username and password.
            These will be sent to the server if it requests authentication.

        :param timeout: Timeouts for the request.
            If a single float, both the connect and read timeout will be set to this value.
            If a tuple, the first value is the connect timeout and the second value is the read timeout.
            If None or 0, no timeout will be set.

        :param allow_redirects:
            If False, do not follow any redirect requests.

        :param proxies: String-pairs mapping protocol to the URL of the proxy.
            Supported protocols are 'ftp', 'http', 'socks5'.

        :param stream: Whether to accept chunked encoding.
            See https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Transfer-Encoding#chunked_encoding

        :param verify: Whether to verify SSL certificates.
            If False, ignore all SSL errors.
            If a string, interpret verify as a path to the CA bundle to verify certificates against.

        :param cert: Client certificate information.
            If a string, interpret cert as a path to a certificate to use for SSL client authentication.
            If a tuple, interpret cert as a (cert, key) pair.

        :param json: JSON data to send in the request body.
            Automatically encodes to bytes and updates Content-Type header.
            Incompatible with the data and files parameters.
        """
        self._request: QNetworkRequest = QNetworkRequest()

        self.method = method
        self.url = url
        self.params = {} if params is None else dict(params)
        self.data = dict(data) if (isinstance(data, Sequence) and not isinstance(data, (bytes, str))) else data
        self.headers = {} if headers is None else dict(headers)
        self.cookies = {} if cookies is None else dict(cookies)
        # self.files = files
        self.auth = auth
        self.timeout = timeout
        self.allow_redirects = allow_redirects
        self.proxies = None if proxies is None else dict(proxies)
        self.stream = stream
        self.verify = verify
        self.cert = cert
        self.json = json

    def __repr__(self) -> str:
        """Representation of the :py:class:`Request` with method and target URL."""
        return f'<Request [{self.method}] ({self.url})>'

    def _prepare_body(self) -> bytes | None:
        content_type = None
        body: bytes | None = None

        if self.data:
            if isinstance(self.data, dict):
                body = encode_url_params(self.data).encode('utf8')
                content_type = 'application/x-www-form-urlencoded'

            elif isinstance(self.data, bytes):
                body = self.data

        elif self.json is not None:
            body = json_dumps(self.json, allow_nan=False).encode('utf8')
            content_type = 'application/json'

        if content_type and 'Content-Type' not in self.headers:
            self.headers['Content-Type'] = content_type

        return body

    def _prepare_headers(self, headers: CaseInsensitiveDict) -> None:
        if self.stream:
            headers['Transfer-Encoding'] = 'chunked'

        if self.cookies:
            headers['Cookie'] = headers

        for name, value in headers.items():
            if name in KNOWN_HEADERS:
                value = _translate_header_value(name, value)
                self._request.setHeader(KNOWN_HEADERS[name][0], value)
                continue

            try:
                encoded_value = bytes(value) if not isinstance(value, str) else value.encode('utf8')
            except TypeError:
                encoded_value = str(value).encode('utf8')

            self._request.setRawHeader(name.encode('utf8'), encoded_value)

    # pylint: disable=compare-to-zero
    def _prepare_response(
            self,
            reply: QNetworkReply,
            finished: _ResponseConsumer | None,
            progress: _ProgressConsumer | None
    ) -> Response:
        _response = Response(self, reply)

        # Put into variables to ignore incorrect known-type errors
        (reply_redirected, reply_finished, reply_downloadProgress, reply_redirectAllowed) = (
            reply.redirected, reply.finished,              # pyright: ignore[reportGeneralTypeIssues]
            reply.downloadProgress, reply.redirectAllowed  # pyright: ignore[reportGeneralTypeIssues]
        )

        if self.allow_redirects:
            reply_redirected.connect(lambda _: reply_redirectAllowed)

        if self.verify is False:
            reply.ignoreSslErrors()

        if finished is not None:
            reply_finished.connect(DeferredCallable(gc_response(finished), _response))

        if progress is not None:
            reply_downloadProgress.connect(DeferredCallable(progress, _response, _extra_pos_args=2))

        if self.timeout:
            # Create connection timeout timer
            # This is for the RESPONSE side of the connection.
            def handle_connection_timeout():
                if not reply.isFinished():
                    reply.abort()

            connection_timeout = int((self.timeout[0] if isinstance(self.timeout, Sequence) else self.timeout) * 1000)
            timer = QTimer(reply)
            timer.setSingleShot(True)
            timer.setInterval(connection_timeout)
            timer.timeout.connect(handle_connection_timeout)  # pyright: ignore[reportGeneralTypeIssues]
            timer.start()

        return _response

    def _prepare_ssl(self) -> None:
        ssl_config = QSslConfiguration.defaultConfiguration()

        if isinstance(self.verify, str):
            ssl_config.setCaCertificates(QSslCertificate.fromPath(self.verify))

        if isinstance(self.cert, str):
            ssl_config.setLocalCertificateChain(QSslCertificate.fromPath(self.cert))
        elif isinstance(self.cert, tuple):
            # cert is a tuple of (cert_path, key_path)
            ssl_config.setLocalCertificateChain(QSslCertificate.fromPath(self.cert[0]))
            ssl_config.setPrivateKey(QSslKey(Path(self.cert[1]).read_bytes(), QSsl.KeyAlgorithm.Rsa))

        self._request.setSslConfiguration(ssl_config)

    def send(self,
             session: NetworkSession,
             wait_until_finished: bool = False,
             finished: _ResponseConsumer | None = None,
             progress: _ProgressConsumer | None = None
             ) -> Response:
        """Send the :py:class:`Request` using the specified :py:class:`NetworkSession`.

        :param session: Session to use.

        :param wait_until_finished: Process the application eventLoop until
            the reply is finished, so when it is returned you can immediately access data.

        :param finished: Callback when the request finishes,
            with reply supplied as an argument.

        :param progress: Callback to update download progress,
            with the reply, received bytes, and total bytes supplied as arguments.

        :return: Response object, which is not guaranteed to be finished.
        :raises ValueError: If proxy attribute is not a valid option.
        """
        request_url = QUrl(self.url)  # Ensure url is of type QUrl
        request_params = query_to_dict(request_url.query()) | self.params  # Update QUrl params with params argument
        request_headers = session.headers | self.headers                   # Use session headers as default headers
        request_cookies = session.cookies | self.cookies                   # Use session cookies as default cookies
        request_data = self._prepare_body()

        if self.cookies:
            request_headers['Cookie'] = request_cookies

        request_url.setQuery(dict_to_query(request_params))
        self._request.setUrl(request_url)

        self._prepare_ssl()
        self._prepare_headers(request_headers)

        if not self.allow_redirects:
            session.manager.setRedirectPolicy(QNetworkRequest.RedirectPolicy.ManualRedirectPolicy)

        if self.proxies is not None:
            for protocol, proxy_url in self.proxies.items():
                proxy_type: QNetworkProxy.ProxyType
                match protocol:
                    case '':
                        proxy_type = QNetworkProxy.ProxyType.NoProxy
                    case 'ftp':
                        proxy_type = QNetworkProxy.ProxyType.FtpCachingProxy
                    case 'http':
                        proxy_type = QNetworkProxy.ProxyType.HttpProxy
                    case 'socks5':
                        proxy_type = QNetworkProxy.ProxyType.Socks5Proxy
                    case other:
                        raise ValueError(f'proxy protocol "{other}" is not supported.')

                proxy_url = QUrl(proxy_url)
                proxy = QNetworkProxy(proxy_type, proxy_url.host(), proxy_url.port())
                session.manager.setProxy(proxy)

        if self.timeout:
            # Set transfer timeout amount
            # This is for the REQUEST side of the connection.
            transfer_timeout = int((self.timeout[1] if isinstance(self.timeout, Sequence) else self.timeout) * 1000)
            self._request.setTransferTimeout(transfer_timeout)

        verb: bytes = self.method.encode('utf8')
        _reply: QNetworkReply = session.manager.sendCustomRequest(self._request, verb, data=request_data)
        response: Response = self._prepare_response(_reply, finished, progress)

        if self.auth:
            session.reply_auth_map[_reply] = self.auth

        if session.manager.redirectPolicy() != session.default_redirect_policy:
            session.manager.setRedirectPolicy(session.default_redirect_policy)

        if wait_until_finished:
            wait_for_reply(_reply)

        return response


class Response:
    """``requests``-like wrapper over a :py:class:`QNetworkReply`."""

    def __init__(self, request: Request, reply: QNetworkReply) -> None:
        """Initialize the :py:class:`Response`."""
        self._data: bytes | None = None
        self._encoding: str | None = None
        self._headers: CaseInsensitiveDict = CaseInsensitiveDict()
        self._reply: QNetworkReply = reply
        self.request: Request = request

    def __del__(self) -> None:
        """Usually the last reference to this :py:class:`Response` is connected to a :py:class:`QNetworkReply` signal.

        So, when the :py:class:`QNetworkReply` is deleted using ``Response.delete()``, ``__del__`` is usually called.
        If this is not the case, and the :py:class:`QNetworkReply` was not yet deleted,
        delete it now to prevent a possible memory leak.
        """
        if Shiboken.isValid(self._reply):
            self._reply.deleteLater()

    def __repr__(self) -> str:
        """Representation of the :py:class:`Response` with its HTTP status code."""
        return f'<Response [{self.code}]>'

    def get_internal_error(self) -> tuple[QNetworkReply.NetworkError, str] | None:
        """Return a tuple containing a ``QNetworkReply.NetworkError`` and its description.

        If the ``QNetworkReply.NetworkError`` is ``NoError``, return ``None``.
        """
        if (err := self._reply.error()) == QNetworkReply.NetworkError.NoError:
            return None

        return err, self._reply.errorString()

    @property
    def code(self) -> int | None:
        """Return the HTTP status code of the :py:class:`Response`.

        ``None`` is returned if the Request is not finished or has been aborted.
        """
        return self._reply.attribute(QNetworkRequest.Attribute.HttpStatusCodeAttribute)

    @property
    def data(self) -> bytes:
        """Return the :py:class:`Response` data as ``bytes``, and cache it for later use."""
        if self._data is None:
            self._data = self._reply.readAll().data()

        return self._data

    @property
    def encoding(self) -> str | None:
        """Return the detected encoding for data, and cache it for later use."""
        if self._encoding is None:
            decoder: QStringDecoder = QStringDecoder.decoderForHtml(self.data)
            self._encoding = str(decoder.name())

        return self._encoding

    @property
    def finished(self) -> bool:
        """Return whether the internal :py:class:`QNetworkReply` is marked as finished."""
        return self._reply.isFinished()

    @property
    def headers(self) -> CaseInsensitiveDict:
        """Return a :py:class:`CaseInsensitiveDict` containing the :py:class:`Response`'s HTTP headers."""
        # Assume that headers don't change after being set.
        if self._headers:
            return self._headers

        # Update with known headers
        for name, (enum_value, _) in KNOWN_HEADERS.lower_items():
            if (value := self._reply.header(enum_value)) is not None:
                self._headers[name] = value

        # Update with raw headers
        for raw_name, raw_value in self._reply.rawHeaderPairs():
            if (name := raw_name.toStdString()) not in self._headers:
                value: bool | int | str
                string_val: str = raw_value.toStdString()

                if string_val.lower() == 'true':
                    value = True

                elif string_val.lower() == 'false':
                    value = False

                elif match := _INT_PATTERN.match(string_val):
                    value = int(match[0])

                else:
                    value = string_val

                self._headers[name] = value

        return self._headers

    @property
    def json(self) -> dict[str, Any]:
        """Return the :py:class:`Response` data as a ``JSON`` object."""
        data: bytes = self.data
        if (encoding := self.encoding) is None:
            encoding = guess_json_utf(data) or 'utf8'

        return json_.loads(data.decode(encoding=encoding))

    @property
    def ok(self) -> bool:
        """Return whether ``self.code`` is a non-error code."""
        if self.code is None:
            return False

        return not is_error_status(self.code)

    @property
    def text(self) -> str:
        """Return the :py:class:`Response` data as a unicode-encoded string."""
        data: bytes = self.data
        encoding: str = self.encoding or 'utf8'

        if not data:
            return ''

        return data.decode(encoding=encoding)

    @property
    def url(self) -> QUrl:
        """Return the URL the :py:class:`Response` is from."""
        return self._reply.url()

    def delete(self) -> None:
        """Delete internal :py:class:`QNetworkReply`.

        If this :py:class:`Response` was connected to a signal which doesn't delete after call (such as progress),
        you will have to call this method before the :py:class:`Response` can be garbage collected.
        """
        self._reply.deleteLater()


_ResponseConsumer: TypeAlias = Callable[[Response], None]
_ProgressConsumer: TypeAlias = Callable[[Response, int, int], None]



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""ComboBox implementations."""
from __future__ import annotations

__all__ = (
    'ComboBox',
    'HistoryComboBox',
    'TranslatableComboBox',
)

from collections.abc import Iterable
from collections.abc import Iterator

from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *

from ...utils import init_objects
from ..aliases import tr
from .paste_line_edit import PasteLineEdit


class ComboBox(QComboBox):
    """Iterable :py:class:`QComboBox`."""

    def __init__(self, *args, **kwargs) -> None:
        """Initialize :py:class:`ComboBox` values."""
        super().__init__(*args, **kwargs)
        self._iter_index = -1

    def __iter__(self) -> Iterator[tuple[str, str]]:
        """Iterate over items and their associated data."""
        self._iter_index = -1
        return self

    def __next__(self) -> tuple[str, str]:
        """Return the next item and its data.

        :return: display text and key, packaged into a tuple.
        :raises StopIteration: When iteration index reaches the last item.
        """
        self._iter_index += 1

        if self._iter_index <= self.count() - 1:
            return self.itemText(self._iter_index), self.itemData(self._iter_index)

        raise StopIteration

    def addItems(self, texts: Iterable[str], *args, **kwargs) -> None:
        """self.addItem for text in texts."""
        for text in texts:
            self.addItem(text, *args, **kwargs)


class HistoryComboBox(ComboBox):
    """Editable :py:class:`ComboBox` acting as a history wrapper over :py:class:`PasteLineEdit`.

    Has no duplicate values.
    """

    line_edit_class = PasteLineEdit

    def __init__(self, *args, **kwargs) -> None:
        """Initialize :py:class:`HistoryComboBox` values."""
        super().__init__(*args, **kwargs)

        init_objects({
            self: {
                'editable': True,
                'duplicatesEnabled': False,
                'lineEdit': self.line_edit_class(parent=self)
            }
        })

    # noinspection PyTypeChecker
    def addItem(self, text: str, *args, **kwargs) -> None:
        """Filter already-present strings from being added using addItem."""
        if (result := self.findText(text, Qt.MatchFlag.MatchFixedString)) != -1:
            self.removeItem(result)

        super().addItem(text, *args, **kwargs)


class TranslatableComboBox(ComboBox):
    """:py:class:`ComboBox` with translatable items."""

    def translate_items(self, *_) -> None:
        """Translate all items with their respective key."""
        items = tuple(key for _, key in self)
        self.clear()
        self.addItems(items)

    # noinspection PyTypeChecker
    def addItem(self, text: str, *args, **kwargs) -> None:
        """Translate strings from being added using addItem."""
        super().addItem(tr(text), *args, userData=text, **kwargs)



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""ExternalTextBrowser implementation."""
from __future__ import annotations

__all__ = (
    'ExternalTextBrowser',
)

from collections import defaultdict
from collections.abc import Callable
from typing import Any

from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *

from ..aliases import app
from ...constants import *
from ...models import DeferredCallable
from ...network import Response


class ExternalTextBrowser(QTextBrowser):
    """:py:class:`QTextBrowser` with ability to map keys to :py:class:`Callable`'s.

    Also supports external image loading and caching.
    """

    remote_image_cache: dict[str, bytes] = {}

    def __init__(self, *args, **kwargs) -> None:
        """Initialize :py:class:`ExternalTextBrowser` values."""
        super().__init__(*args, **kwargs)
        self.anchorClicked.connect(  # pyright: ignore[reportGeneralTypeIssues]
            lambda url: self.scrollToAnchor(url.toDisplayString().lstrip('#'))
        )

        # default factory producing empty DeferredCallables
        self.key_callable_map: defaultdict[int, Callable] = defaultdict(DeferredCallable)
        self.inject_markdown_anchors: bool = True
        self.cached_text: str = ''
        self.cached_type: str = ''

    def hot_reload(self) -> None:
        """Reload cached text using the cached type's function."""
        # Remember scroll position
        scroll = self.verticalScrollBar().sliderPosition()

        # Reload all resources and replace text with cached text
        self.clear()
        match self.cached_type:
            case 'markdown':
                self.setMarkdown(self.cached_text)
            case 'html':
                self.setHtml(self.cached_text)
            case 'text':
                self.setPlainText(self.cached_text)
        self.verticalScrollBar().setSliderPosition(scroll)

    def set_hot_reloadable_text(self, text: str, text_type: str) -> None:
        """Set text that is designated to be hot-reloadable.

        Injects html anchors to images to allow clickable images.
        Injects html anchors to markdown headers to allow relative linking.
        """
        # Inject html anchors in images and headers
        if text_type == 'markdown':
            lines: list[str] = text.splitlines()
            lines_new: list[str] = []
            ref_links: dict[str, tuple[str, str | None]] = {
                match['label']: (match['url'], match['description'])
                for match in MARKDOWN_REF_LINK_PATTERN.finditer(text)
            }

            for i, line in enumerate(lines):

                # Replace image links with clickable anchor images
                if match := MARKDOWN_IMG_LINK_PATTERN.match(line):
                    # Get ref_link or normal link for anchor href
                    stripped_groups = tuple(group.strip('()[]') for group in match.groups())
                    alt: str = stripped_groups[0]
                    src: str = stripped_groups[1]
                    link: str = stripped_groups[2]

                    # Get reference link data if applicable
                    link_is_ref: bool = match[0][2].startswith('[')
                    ref_url: str | None = ref_links.get(link, (None, None))[0]
                    ref_desc: str | None = ref_links.get(link, (None, None))[1]

                    # Set href and title
                    href: str
                    title: str

                    if link_is_ref:
                        href = ref_url if ref_url is not None else link
                        title = ref_desc if ref_desc is not None else alt
                    else:
                        href = link
                        title = alt

                    lines_new.append(f'<a href="{href}"><img src="{src}" alt="{alt}" title="{title}"/></a>')
                    continue

                # Find headers and add relative anchors
                if self.inject_markdown_anchors:
                    has_hashtag: bool = line.strip().startswith('#')
                    has_underline: bool = bool(line) and (i < len(lines) - 1) and any(
                        # Line must end and begin with the respective underline
                        lines[i + 1].strip().startswith(line_char) and
                        lines[i + 1].strip().endswith(line_char) for line_char in ('-', '=')
                    )

                    # If line has a header marker, append it with an anchor tag and add it to the new list.
                    # Otherwise, append the unchanged line to the new list.
                    if has_hashtag or has_underline:
                        simple_str: str = line.strip().replace(' ', '-').replace(':', '').lstrip('#').strip('-').lower()

                        lines_new.append(
                            f'{line} <a name="{simple_str}" href="#{simple_str}" '
                            f'style="font-size: 1px; color: transparent">§</a>'
                        )
                        continue

                lines_new.append(line)

            text = '\n'.join(lines_new)

        self.cached_text = text
        self.cached_type = text_type
        self.hot_reload()

    def loadResource(self, resource_type: int, url: QUrl, **kwargs) -> Any:
        """Load a resource from an url.

        If resource type is an image and the url is external, download it using requests and cache it.
        """
        # See: https://github.com/Cubicpath/HaloInfiniteGetter/pull/30#issuecomment-1279562460
        # QTextDocument.ResourceType.ImageResource = 0x2
        if resource_type == 0x2 and not url.isLocalFile():
            image: QImage = QImage()
            if (url_string := url.toDisplayString()) not in self.remote_image_cache:
                # Add placeholder bytes to show the url is being downloaded.
                # Otherwise, unneeded replies are sent out and cause application stutter.
                self.remote_image_cache[url_string] = b''

                def handle_reply(reply: Response):
                    image.loadFromData(reply.data)
                    self.remote_image_cache[url_string] = reply.data

                    if self.cached_type:
                        self.hot_reload()

                app().session.get(url, finished=handle_reply)
            elif cached_data := self.remote_image_cache[url_string]:
                image.loadFromData(cached_data)
            return image

        return super().loadResource(resource_type, url, **kwargs)

    def setLineWrapMode(self, mode: int | QTextEdit.LineWrapMode) -> None:
        """Set the line wrap mode. Allows use of ints."""
        super().setLineWrapMode(QTextEdit.LineWrapMode(mode))

    def connect_key_to(self, key: Qt.Key, func: Callable) -> None:
        """Connect a :py:class:`Callable` to a key press."""
        self.key_callable_map[int(key)] = func

    def keyPressEvent(self, event: QKeyEvent) -> None:
        """Execute :py:class:`Callable` mapped to the key press."""
        super().keyPressEvent(event)
        self.key_callable_map[event.key()]()
        event.accept()



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""Column menu implementation."""
from __future__ import annotations

__all__ = (
    'ColumnContextMenu',
)

from PySide6.QtGui import *
from PySide6.QtWidgets import *

from ...models import DeferredCallable
from ...utils import add_menu_items
from ...utils import init_objects
from ..aliases import app


class ColumnContextMenu(QMenu):
    """A general context menu for hiding/unhiding columns in a :py:class:`QTreeView`."""

    def __init__(self, parent: QTreeView, disabled_indices: set[int] | None = None):
        """Create a new :py:class:`ColumnContextMenu` for the given :py:class:`QTreeView`.

        You can specifically disable any header indices you want.
        """
        super().__init__(parent)

        disabled_indices = set() if disabled_indices is None else disabled_indices

        icons = (
            # Not Hidden
            app().get_theme_icon('checkbox_checked') or
            self.style().standardIcon(QStyle.StandardPixmap.SP_DialogApplyButton),

            # Hidden
            app().get_theme_icon('checkbox_unchecked') or
            self.style().standardIcon(QStyle.StandardPixmap.SP_DialogCancelButton),
        )

        items: list[str | QAction] = ['Columns']
        for i, hidden in enumerate([parent.isColumnHidden(count) for count in range(parent.model().columnCount())]):
            init_objects({
                (action := QAction(self)): {
                    'disabled': i in disabled_indices,
                    'text': parent.model().headerData(i, Qt.Orientation.Horizontal),
                    'icon': icons[hidden],
                    'triggered': DeferredCallable(parent.setColumnHidden, i, not hidden)
                }
            })

            items.append(action)

        add_menu_items(self, items)



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""Module for the main application classes."""
from __future__ import annotations

__all__ = (
    'GetterApp',
    'Theme',
)

import json
import shutil
import subprocess
import sys
from collections import defaultdict
from collections.abc import Callable
from collections.abc import Sequence
from pathlib import Path
from typing import Any
from typing import Final
from typing import NamedTuple
from typing import TypeAlias

import toml
from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *

from .._version import __version__
from ..constants import *
from ..events import EventBus
from ..lang import Translator
from ..models import DeferredCallable
from ..models import DistributedCallable
from ..models import Singleton
from ..network import Client
from ..network import NetworkSession
from ..network import Response
from ..network import VersionChecker
from ..tomlfile import CommentValue
from ..tomlfile import PathTomlDecoder
from ..tomlfile import PathTomlEncoder
from ..tomlfile import TomlEvents
from ..tomlfile import TomlFile
from ..tomlfile import TomlValue
from ..utils import has_package
from ..utils import hide_windows_file
from ..utils import http_code_map
from ..utils import icon_from_bytes
from ..utils import set_or_swap_icon

_DEFAULTS_FILE: Final[Path] = HI_RESOURCE_PATH / 'default_settings.toml'
_LAUNCHED_FILE: Final[Path] = HI_CONFIG_PATH / '.LAUNCHED'
_SETTINGS_FILE: Final[Path] = HI_CONFIG_PATH / 'settings.toml'

_ButtonsWithRoles: TypeAlias = (
    Sequence[tuple[QAbstractButton, QMessageBox.ButtonRole] | QMessageBox.StandardButton] |
    QMessageBox.StandardButton
)


class _DialogResponse(NamedTuple):
    """Response object for GetterApp.show_dialog()."""

    button: QAbstractButton | QMessageBox.StandardButton = QMessageBox.StandardButton.NoButton
    role: QMessageBox.ButtonRole = QMessageBox.ButtonRole.InvalidRole


class Theme(NamedTuple):
    """Object containing data about a Theme."""

    id: str
    style: str
    display_name: str


class GetterApp(Singleton, QApplication):
    """The main HaloInfiniteGetter PySide application that runs in the background and manages the process.

    :py:class:`GetterApp` is a singleton and can be accessed via the class using the
    GetterApp.instance() class method or the app() function.
    """

    _singleton_base_type = QApplication
    _singleton_check_ref = False
    updateTranslations = Signal()

    # PyCharm detects dict literals in __init__ as a dict[str, EventBus[TomlEvent]], for no explicable reason.
    # noinspection PyTypeChecker
    def __init__(self, *args: str) -> None:
        """Create a new app with the given arguments and settings."""
        super().__init__(list(args))  # Despite documentation saying it takes in a Sequence[str], it only accepts lists
        self.setApplicationName(HI_PACKAGE_NAME)
        self.setApplicationVersion(__version__)

        self._first_launch: bool = not _LAUNCHED_FILE.is_file()  # Check if launched marker exists
        self._legacy_style: str = self.styleSheet()              # Set legacy style before it is overridden
        self._thread_pool: QThreadPool = QThreadPool.globalInstance()

        self._setting_defaults: dict[str, TomlValue | CommentValue] = toml.loads(
            _DEFAULTS_FILE.read_text(encoding='utf8'), decoder=PathTomlDecoder()
        )
        self._registered_translations: DistributedCallable[set, None, None] = DistributedCallable(set())
        self._windows: dict[str, QWidget] = {}

        # Create all files/directories that are needed for the app to run
        self._create_paths()

        # Must have themes up before load_env
        self.icon_store: defaultdict[str, QIcon] = defaultdict(QIcon)  # Null icon generator
        self.session: NetworkSession = NetworkSession(self)
        self.settings: TomlFile = TomlFile(_SETTINGS_FILE, default=self._setting_defaults)  # type: ignore
        self.themes: dict[str, Theme] = {}
        self.theme_index_map: dict[str, int] = {}
        self.translator: Translator
        self.version_checker: VersionChecker = VersionChecker(self)

        # Correct malformed language tag with default language selected by Translator
        try:
            self.translator = Translator(self.settings['language'])  # type: ignore
        except ValueError:
            self.translator = Translator()
            self.settings['language'] = self.translator.language.tag
            self.settings.save()

        # Load resources from disk
        self.load_themes()  # Depends on icon_store, settings, themes, theme_index_map
        self.load_icons()   # Depends on icon_store, session

        # Register formats for use in shutil
        self._register_archive_formats()

        # Must load client last, but before windows
        self.load_env(verbose=True)
        self.client = Client(self)
        self._connect_events()

        # Setup window instances
        self._create_windows()
        self.updateTranslations.emit()

        # Check version after all widgets are created to listen for newerVersion signal
        self.version_checker.check_version()

    @property
    def first_launch(self) -> bool:
        """Return whether this is the first launch of the application.

        This is determined by checking if the .LAUNCHED file exists in the user's config folder.
        """
        return self._first_launch

    @property
    def windows(self) -> dict[str, QWidget]:
        """Return a copy of the self._windows dictionary."""
        return self._windows.copy()

    def _create_paths(self) -> None:
        """Create files and directories if they do not exist."""
        for dir_path in (HI_CACHE_PATH, HI_WEB_DUMP_PATH, HI_CONFIG_PATH):
            if not dir_path.is_dir():
                dir_path.mkdir(parents=True)

        if self.first_launch:
            # Create first-launch marker
            _LAUNCHED_FILE.touch()
            hide_windows_file(_LAUNCHED_FILE)

        if not _SETTINGS_FILE.is_file():
            # Write default_settings to user's SETTINGS_FILE
            with _SETTINGS_FILE.open(mode='w', encoding='utf8') as file:
                toml.dump(self._setting_defaults, file, encoder=PathTomlEncoder())

    def _create_windows(self) -> None:
        """Create window instances."""
        from .windows import AppWindow
        from .windows import ChangelogViewer
        from .windows import LicenseViewer
        from .windows import ReadmeViewer
        from .windows import ScanSelectorDialog
        from .windows import SettingsWindow

        SettingsWindow.create(QSize(420, 600))
        AppWindow.create(QSize(
            # Size to use, with a minimum of 100x100
            max(self.settings['gui/window/x_size'], 100),  # type: ignore
            max(self.settings['gui/window/y_size'], 100)   # type: ignore
        ))

        self._windows['changelog_viewer'] = ChangelogViewer()
        self._windows['license_viewer'] = LicenseViewer()
        self._windows['readme_viewer'] = ReadmeViewer()
        self._windows['scan_selector'] = ScanSelectorDialog()
        self._windows['settings'] = SettingsWindow.instance()  # type: ignore
        self._windows['app'] = AppWindow.instance()            # type: ignore

    def _translate_http_code_map(self) -> None:
        """Translate the HTTP code map to the current language."""
        for code in (400, 401, 403, 404, 405, 406):
            http_code_map[code] = (http_code_map[code][0], self.translator(f'network.http.codes.{code}.description'))

    def _connect_events(self) -> None:
        """Register callables to events."""
        EventBus['settings'] = self.settings.event_bus
        EventBus['settings'].subscribe(
            DeferredCallable(self.load_themes),
            TomlEvents.Import)

        EventBus['settings'].subscribe(
            DeferredCallable(self.updateTranslations.emit),
            TomlEvents.Import)

        EventBus['settings'].subscribe(
            DeferredCallable(self.update_stylesheet),
            TomlEvents.Set, event_predicate=lambda e: e.key == 'gui/themes/selected')

        if not self.settings['ignore_updates']:
            self.version_checker.newerVersion.connect(self._upgrade_version_dialog)

        self.updateTranslations.connect(lambda: setattr(self.translator, 'language', self.settings['language']))
        self.updateTranslations.connect(lambda: self.setApplicationDisplayName(self.translator('app.name')))
        self.updateTranslations.connect(self._registered_translations)
        self.updateTranslations.connect(self._translate_http_code_map)

    def _register_archive_formats(self) -> None:
        if not has_package('py7zr'):
            self.missing_package_dialog('py7zr', 'Importing/Exporting 7Zip Archives')
        if not has_package('py7zr'):  # During the dialog, the package may be dynamically installed by user.
            return

        from py7zr import pack_7zarchive
        from py7zr import unpack_7zarchive

        # Register .7z archive support
        shutil.register_archive_format('7z', pack_7zarchive, description='7Zip Archive File')
        shutil.register_unpack_format('7z', ['7z'], unpack_7zarchive, description='7Zip Archive File')

    def init_translations(self, translation_calls: dict[Callable, str | tuple[Any, ...]]) -> None:
        """Initialize the translation of all objects.

        Register functions to call with their respective translation keys.
        This is used to translate everything in the GUI.
        """
        for func, args in translation_calls.items():
            if not isinstance(args, tuple):
                args = (args,)

            # Call the function with the deferred translation of the given key.
            translate = DeferredCallable(func, DeferredCallable(self.translator, *args))

            # Register the object for dynamic translation
            self._registered_translations.callables.add(translate)

    def update_stylesheet(self) -> None:
        """Set the application stylesheet to the one currently selected in settings."""
        try:
            self.setStyleSheet(self.themes[self.settings['gui/themes/selected']].style)  # type: ignore
        except KeyError:
            self.settings['gui/themes/selected'] = 'legacy'
            self.setStyleSheet(self._legacy_style)

    def show_dialog(self, key: str, parent: QWidget | None = None,
                    buttons: _ButtonsWithRoles | None = None,
                    default_button: QAbstractButton | QMessageBox.StandardButton | None = None,
                    title_args: Sequence | None = None,
                    description_args: Sequence | None = None,
                    details_text: str = '') -> _DialogResponse:
        r"""Show a dialog. This is a wrapper around QMessageBox creation.

        The type of dialog icon depends on the key's first section.
        The following sections are supported::
            - 'about':      -> QMessageBox.about
            - 'questions'   -> QMessageBox.Question
            - 'information' -> QMessageBox.Information
            - 'warnings'    -> QMessageBox.Warning
            - 'errors'      -> QMessageBox.Critical

        The dialog title and description are determined from the
        "title" and "description" child sections of the given key.

        Example with given key as "questions.key"::
            "questions.key.title": "Question Title"\n
            "questions.key.description": "Question Description"

        WARNING: If a StandardButton is clicked, the button returned is NOT a StandardButton enum, but a QPushButton.

        :param key: The translation key to use for the dialog.
        :param parent: The parent widget to use for the dialog. If not supplied, a dummy widget is temporarily created.
        :param buttons: The buttons to use for the dialog. If button is not a StandardButton,
          it should be a tuple containing the button and its role.
        :param default_button: The default button to use for the dialog.
        :param title_args: The translation arguments used to format the title.
        :param description_args: The translation arguments used to format the description.
        :param details_text: If provided, a new 'Show Details...' button will show this text in the dialog when pressed.
        :return: The button that was clicked, as well as its role. None if the key's first section is "about".
        :raises TypeError: If default_button is not a QPushButton or a QStandardButton.
        """
        using_dummy_widget: bool = False
        if parent is None:
            parent = QWidget(None)
            using_dummy_widget = True

        title_args = () if title_args is None else title_args
        description_args = () if description_args is None else description_args

        icon: QMessageBox.Icon
        first_section: str = key.split('.')[0]
        match first_section:
            case 'questions':
                icon = QMessageBox.Icon.Question
            case 'information':
                icon = QMessageBox.Icon.Information
            case 'warnings':
                icon = QMessageBox.Icon.Warning
            case 'errors':
                icon = QMessageBox.Icon.Critical
            case _:
                icon = QMessageBox.Icon.NoIcon

        title_text: str = self.translator(f'{key}.title', *title_args)
        description_text: str = self.translator(f'{key}.description', *description_args)

        # Create custom QMessageBox
        msg_box = QMessageBox(icon, title_text, description_text, parent=parent)
        msg_box.setTextInteractionFlags(Qt.TextInteractionFlag.TextBrowserInteraction)

        if details_text:
            msg_box.setDetailedText(details_text)

        # If no icon is selected, it follows the precedence of QMessagebox.about(). That being:
        # 1. It prefers parent.icon() if that exists.
        # 2. If not, it tries the top-level widget containing parent.
        # 3. If that fails, it tries the PySide6.QtWidgets.QApplication.activeWindow()
        # 4. As a last resort it uses the Information icon.
        if icon is QMessageBox.Icon.NoIcon:
            for widget in (parent, parent.topLevelWidget(), QApplication.activeWindow()):
                if not (about_icon := widget.windowIcon()).isNull():
                    msg_box.setIconPixmap(about_icon.pixmap(999))
                    break
            else:
                msg_box.setIcon(QMessageBox.Icon.Information)

        # Assemble the buttons in the correct order.
        standard_buttons = None
        if buttons is not None:
            if isinstance(buttons, Sequence):
                # If the button is not a tuple, assume it's a QMessageBox.StandardButton.
                # Build a StandardButtons from all StandardButton objects in buttons.
                for button in buttons:
                    if isinstance(button, tuple):
                        msg_box.addButton(button[0], button[1])
                    elif standard_buttons is None:
                        standard_buttons = button
                    else:
                        standard_buttons |= button
            else:
                # If the buttons is not a sequence, assume it's QMessageBox.StandardButtons.
                standard_buttons = buttons

        if standard_buttons:
            msg_box.setStandardButtons(standard_buttons)

        if default_button is not None:
            if not isinstance(default_button, (QPushButton, QMessageBox.StandardButton)):
                raise TypeError(f'Default button cannot be of type: {type(default_button)}')

            msg_box.setDefaultButton(default_button)

        # Add buttonClicked result to list to extract as variable.
        # This works as msg_box.exec() blocks the event loop until finished.
        msg_box.buttonClicked.connect((result := []).append)  # pyright: ignore[reportGeneralTypeIssues]
        msg_box.exec()

        # Get data from result to create a _DialogResponse from
        response_kwargs = {}
        if result:
            response_kwargs['button'] = result[0]
            response_kwargs['role'] = msg_box.buttonRole(result[0])

        # If parent wasn't specified, delete the dummy widget.
        # This will also delete any button objects attached to the QMessageBox.
        if using_dummy_widget:
            # noinspection PyUnboundLocalVariable
            parent.deleteLater()

        return _DialogResponse(**response_kwargs)

    def missing_package_dialog(self, package: str, reason: str | None = None, parent: QWidget | None = None) -> None:
        """Show a dialog informing the user that a package is missing and asks to install said package.

        If a user presses the "Install" button, the package is installed.

        :param package: The name of the package that is missing.
        :param reason: The reason why the package is attempting to be used.
        :param parent: The parent widget to use for the dialog. If not supplied, a dummy widget is temporarily created.
        """
        exec_path = Path(sys.executable)

        install_button = QPushButton(
            self.get_theme_icon('dialog_ok'),
            self.translator('errors.missing_package.install')
        )

        consent_to_install: bool = self.show_dialog(
            'errors.missing_package', parent, (
                (install_button, QMessageBox.ButtonRole.AcceptRole),
                QMessageBox.StandardButton.Cancel
            ),
            default_button=QMessageBox.StandardButton.Cancel,
            description_args=(package, reason, exec_path)
        ).role == QMessageBox.ButtonRole.AcceptRole

        if consent_to_install:
            try:
                # Install the package, Path(sys.executable) contains 0 user input.
                subprocess.run(  # nosec B603:subprocess_without_shell_equals_true
                    (exec_path, '-m', 'pip', 'install', package),
                    check=True
                )
            except (OSError, subprocess.SubprocessError) as e:
                self.show_dialog(
                    'errors.package_install_failure', parent,
                    description_args=(package, e)
                )
            else:
                self.show_dialog(
                    'information.package_installed', parent,
                    description_args=(package,)
                )

    def _upgrade_version_dialog(self, _, version: str) -> None:
        ignore_button = QPushButton(
            self.translator('information.upgrade_version.ignore')
        )

        upgrade_button = QPushButton(
            self.get_theme_icon('dialog_ok'),
            self.translator('information.upgrade_version.upgrade')
        )

        match self.show_dialog(
            'information.upgrade_version', None, (
                (upgrade_button, QMessageBox.ButtonRole.YesRole),
                (ignore_button, QMessageBox.ButtonRole.NoRole),
                QMessageBox.StandardButton.Cancel
            ),
            default_button=QMessageBox.StandardButton.Cancel,
            description_args=(version, __version__)
        ).role:
            case QMessageBox.ButtonRole.YesRole:
                QProcess.startDetached(sys.executable, arguments=[
                    str((HI_RESOURCE_PATH / 'scripts/upgrade_version.py').absolute().resolve(True)),
                    f'{HI_PACKAGE_NAME.replace("_", "-")}', HI_PACKAGE_NAME
                ])
            case QMessageBox.ButtonRole.NoRole:
                self.version_checker.newerVersion.disconnect(self._upgrade_version_dialog)
                self.settings['ignore_updates'] = True
                self.settings.save()

    def load_env(self, verbose: bool = True) -> None:
        """Load environment variables from .env file."""
        if not has_package('python-dotenv'):
            self.missing_package_dialog('python-dotenv', 'Loading environment variables')
        if not has_package('python-dotenv'):  # During the dialog, the package may be dynamically installed by user.
            return

        from dotenv import load_dotenv
        load_dotenv(verbose=verbose)

    def load_icons(self) -> None:
        """Load all icons needed for the application.

        Fetch locally stored icons from the HI_RESOURCE_PATH/icons directory

        Asynchronously fetch externally stored icons from urls defined in HI_RESOURCE_PATH/external_icons.json
        """
        # Load locally stored icons
        self.icon_store.update({
            filename.stem: QIcon(str(filename)) for
            filename in (HI_RESOURCE_PATH / 'icons').iterdir() if filename.is_file()
        })

        # Load external icon links
        external_icon_links: dict[str, str] = json.loads(
            (HI_RESOURCE_PATH / 'external_icons.json').read_text(encoding='utf8')
        )

        # Load externally stored icons
        # pylint: disable=cell-var-from-loop
        for key, url in external_icon_links.items():
            # Create a new handler for every key being requested.
            def handle_reply(reply: Response):
                icon = icon_from_bytes(reply.data)
                set_or_swap_icon(self.icon_store, key, icon)

            self.session.get(url, finished=handle_reply)

        # Set the default icon for all windows.
        self.setWindowIcon(self.icon_store['hi'])

    def warn_for(self, warning_key: str) -> None:
        """One-time warning for users of application.

        :param warning_key: language key to show dialog for and to write in the ``.warned`` config file.
        """
        with (HI_CONFIG_PATH / '.warned').open(mode='a+', encoding='utf8') as warned_file:
            warned_file.seek(0)
            acknowledged_warnings: set[str] = {line.strip() for line in warned_file.readlines()}

            if warning_key not in acknowledged_warnings:
                self.show_dialog(warning_key)
                warned_file.write(f'{warning_key}\n')

    def get_theme_icon(self, icon: str) -> QIcon:
        """Return the icon for the given theme.

        :param icon: Icon name for given theme.
        :return: QIcon for the given theme or a null QIcon if not found. Null icons are falsy.
        """
        current_theme = self.settings['gui/themes/selected']
        return self.icon_store[f'hi_theme+{current_theme}+{icon}']

    def add_theme(self, theme: Theme) -> None:
        """Add a theme to the application.

        Overwrites previous theme if the ids are the same.
        """
        self.themes[theme.id] = theme

    def load_themes(self) -> None:
        """Load all theme locations from settings and store them in self.themes.

        Also set current theme from settings.

        :raises ValueError: If gui/themes/{theme}/path setting is not a string.
        """
        self.add_theme(Theme('legacy', self._legacy_style, 'Legacy (Default Qt)'))

        try:
            themes: dict[str, TomlValue] = self.settings['gui/themes']  # type: ignore
        except KeyError:
            self.settings['gui/themes'] = themes = {}

        for id, theme in themes.items():
            if not isinstance(theme, dict):
                continue

            if isinstance((path_ := theme['path']), CommentValue):
                path_ = path_.val

            if not isinstance(path_, str):
                raise ValueError(f'gui/themes/{theme}/path is not a string value.')

            # Translate builtin theme locations
            if path_.startswith('builtin::'):
                path_ = HI_RESOURCE_PATH / f'themes/{path_.removeprefix("builtin::")}'

            # Ensure path is a Path value that exists
            if (path := Path(path_)).is_dir():
                QDir.addSearchPath(f'hi_theme+{id}', path)
                theme_attrs = {
                    'id': id,
                    'style': '',
                    'display_name': theme.get('display_name') or id
                }

                for theme_resource in path.iterdir():
                    if theme_resource.is_file():
                        if theme_resource.name == 'stylesheet.qss':
                            # Load stylesheet file
                            theme_attrs['style'] = theme_resource.read_text(encoding='utf8')
                        elif theme_resource.suffix.lstrip('.') in SUPPORTED_IMAGE_EXTENSIONS:
                            # Load all images in the theme directory into the icon store.
                            theme_key = f'hi_theme+{id}+{theme_resource.stem}'
                            self.icon_store[theme_key] = QIcon(str(theme_resource.resolve()))

                self.add_theme(Theme(**theme_attrs))

        # noinspection PyUnresolvedReferences
        self.theme_index_map = {theme_id: i for i, theme_id in enumerate(theme.id for theme in self.sorted_themes())}
        self.update_stylesheet()

    def sorted_themes(self) -> list[Theme]:
        """List of themes sorted by their display name."""
        return sorted(self.themes.values(), key=lambda theme: theme.display_name)

    def start_worker(self, runnable: Callable | QRunnable, priority: int = 0) -> None:
        """Start a runnable from the application :py:class:`QThreadPool`."""
        self._thread_pool.start(runnable, priority)



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""Convenience alias functions."""
from __future__ import annotations

__all__ = (
    'app',
    'tr',
)

from typing import Any

from .app import GetterApp


def app() -> GetterApp:
    """Return :py:class:`GetterApp`.instance()."""
    return GetterApp.instance()


def tr(key: str, *args: Any, **kwargs: Any) -> str:
    """Alias for app().translator().

    :param key: Translation keys to translate.
    :param args: Arguments to format key with.
    :keyword default: Default value to return if key is not found.
    :return: Translated text.
    """
    return app().translator(key, *args, **kwargs)



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""Help menu implementation."""
from __future__ import annotations

__all__ = (
    'HelpContextMenu',
    'package_versions',
)

import sys
from platform import platform

from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *

from ..._version import __version_info__
from ...constants import *
from ...models import DeferredCallable
from ...utils import add_menu_items
from ...utils import init_objects
from ...utils import current_requirement_versions
from ..aliases import app
from ..aliases import tr


def package_versions() -> str:
    """Generate the package version list for use in the "About" message.

    :return: Package version list separated by newlines.
    """
    return '\n'.join(
        tr('about.app.package_version', package, version) for
        package, version in current_requirement_versions(HI_PACKAGE_NAME, include_extras=True).items()
    )


class HelpContextMenu(QMenu):
    """Context menu that shows actions to help the user."""

    def __init__(self, parent) -> None:
        """Create a new :py:class:`HelpContextMenu`."""
        super().__init__(parent)

        init_objects({
            (github_view := QAction(self)): {
                'text': tr('gui.menus.help.github'),
                'icon': app().icon_store['github'],
                'triggered': DeferredCallable(
                    QDesktopServices.openUrl,
                    QUrl('https://github.com/Cubicpath/HaloInfiniteGetter/')
                )
            },

            (create_issue := QAction(self)): {
                'text': tr('gui.menus.help.issue'),
                'icon': app().icon_store['github'],
                'triggered': DeferredCallable(
                    QDesktopServices.openUrl,
                    QUrl('https://github.com/Cubicpath/HaloInfiniteGetter/issues/new/choose')
                )
            },

            (about_view := QAction(self)): {
                'text': tr('gui.menus.help.about'),
                'icon': (app().get_theme_icon('message_question') or
                         app().icon_store['about']),
                'triggered': self.open_about
            },

            (about_qt_view := QAction(self)): {
                'text': tr('gui.menus.help.about_qt'),
                'icon': (app().get_theme_icon('message_question') or
                         app().icon_store['about']),
                'triggered': DeferredCallable(
                    QMessageBox(self).aboutQt, self, tr('about.qt.title')
                )
            },

            (changelog := QAction(self)): {
                'text': tr('gui.menus.help.changelog'),
                'icon': (app().get_theme_icon('message_information') or
                         self.style().standardIcon(QStyle.StandardPixmap.SP_DialogApplyButton)),
                'triggered': lambda: app().windows['changelog_viewer'].show()
            },

            (license_view := QAction(self)): {
                'text': tr('gui.menus.help.license'),
                'icon': app().icon_store['copyright'],
                'triggered': lambda: app().windows['license_viewer'].show()
            },

            (readme := QAction(self)): {
                'text': tr('gui.menus.help.readme'),
                'icon': (app().get_theme_icon('message_information') or
                         self.style().standardIcon(QStyle.StandardPixmap.SP_DialogApplyButton)),
                'triggered': lambda: app().windows['readme_viewer'].show()
            }
        })

        add_menu_items(self, [
            'Github', github_view, create_issue,
            'Docs', changelog, readme,
            'About', about_view, about_qt_view, license_view
        ])

    # pylint: disable=not-an-iterable
    def open_about(self) -> None:
        """Open the application's about section."""
        app().show_dialog(
            'about.app', self,
            description_args=(
                *__version_info__, platform(), sys.implementation.name,
                sys.version.split('[', maxsplit=1)[0],
                package_versions()
            )
        )



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""Module used for TOML configurations."""
from __future__ import annotations

__all__ = (
    'CommentValue',
    'PathTomlDecoder',
    'PathTomlEncoder',
    'TomlFile',
    'TomlEvents',
    'TomlValue',
)

import warnings
from collections.abc import Callable
from collections.abc import MutableMapping
from pathlib import Path
from pathlib import PurePath
from typing import Any
from typing import Final
from typing import TypeAlias
from typing import TypeVar

import toml
from toml.decoder import CommentValue as _CommentValue

from .events import Event
from .events import EventBus

_COMMENT_PREFIX: Final[str] = '# '
_SPECIAL_PATH_PREFIX: Final[str] = '$PATH$|'

TomlValue: TypeAlias = dict[str, 'TomlValue'] | list['TomlValue'] | float | int | str | bool | PurePath
"""Represents a possible TOML value, with :py:class:`dict` being a Table, and :py:class:`list` being an Array."""


_TT = TypeVar('_TT', bound=TomlValue)  # Bound to TomlValue.


class _MetaCommentValue(type):
    """Overrides instance checks so that :py:class:`_CommentValue` is accepted as a :py:class:`CommentValue`."""

    def __instancecheck__(cls, instance: Any) -> bool:
        return isinstance(instance, _CommentValue)


class CommentValue(_CommentValue, metaclass=_MetaCommentValue):
    """Properly typed version of :py:class:`_CommentValue`."""

    def __init__(self, val: TomlValue, comment: str | None = None,
                 new_line: bool = False, _dict_type: type[MutableMapping] = dict) -> None:
        """Properly annotates types for :py:class:`_CommentValue`."""
        comment = f'{_COMMENT_PREFIX}{comment}' if comment is not None else ''
        separator: str = '\n' if new_line else ' '
        super().__init__(val, comment, new_line, _dict_type)

        self.val: TomlValue = val
        self.comment: str = separator + comment
        self._dict: type[MutableMapping] = _dict_type

    def __getitem__(self, key: str) -> TomlValue:
        """Proxy for :py:class:`MutableMapping`.__getitem__ if ``self.val`` is a :py:class:`MutableMapping`.

        :raises TypeError: If self.val is not a MutableMapping.
        """
        if not isinstance(self.val, MutableMapping):
            raise TypeError(f'Cannot access item for non-{self._dict} value.')
        return self.val[key]

    def __setitem__(self, key: str, value: TomlValue):
        """Proxy for :py:class:`MutableMapping`.__setitem__ if ``self.val`` is a :py:class:`MutableMapping`.

        :raises TypeError: If self.val is not a MutableMapping.
        """
        if not isinstance(self.val, MutableMapping):
            raise TypeError(f'Cannot assign item for non-{self._dict} value.')
        self.val[key] = value

    @classmethod
    def from_comment_val(cls, _comment_val: _CommentValue) -> CommentValue:
        """Make a :py:class:`CommentValue` from a :py:class:`_CommentValue`."""
        return cls(_comment_val.val, _comment_val.comment)

    def dump(self, dump_value_func: Callable[[TomlValue], str]) -> str:
        """Dump :py:class:`CommentValue` into a str."""
        ret_str: str = dump_value_func(self.val)
        if isinstance(self.val, self._dict):
            return self.comment + '\n' + str(ret_str)
        return str(ret_str) + self.comment


class TomlEvents:
    """Namespace for all events relating to :py:class:`TomlFile` objects."""

    class TomlEvent(Event):
        """Generic event for TomlFiles."""

        __slots__ = ('toml_file',)

        def __init__(self, toml_file: TomlFile) -> None:
            """Create a new :py:class:`TomlEvents.TomlEvent` event with the given file."""
            self.toml_file = toml_file

    class File(TomlEvent):
        """Accessing a File on disk."""

        __slots__ = ('path',)

        def __init__(self, toml_file: TomlFile, path: Path) -> None:
            """Create a new :py:class:`TomlEvents.File` event with the given file and path."""
            super().__init__(toml_file=toml_file)
            self.path = path

    class Import(File):
        """Loading a TomlFile."""

        __slots__ = ()

    class Export(File):
        """Exporting a TomlFile to disk."""

        __slots__ = ()

    class KeyAccess(TomlEvent):
        """A key's value is accessed."""

        __slots__ = ('key',)

        def __init__(self, toml_file: TomlFile, key: str) -> None:
            """Create a new :py:class:`TomlEvents.KeyAccess` event with the given file and key."""
            super().__init__(toml_file=toml_file)
            self.key: str = key

    class Get(KeyAccess):
        """Value is given."""

        __slots__ = ('value',)

        def __init__(self, toml_file: TomlFile, key: str, value: TomlValue) -> None:
            """Create a new :py:class:`TomlEvents.Get` event with the given file, key, and value."""
            super().__init__(toml_file=toml_file, key=key)
            self.value = value

    class Set(KeyAccess):
        """Value is set."""

        __slots__ = ('old', 'new')

        def __init__(self, toml_file: TomlFile, key: str, old: TomlValue | None, new: TomlValue) -> None:
            """Create a new :py:class:`TomlEvents.Set` event with the given file, key, old value, and new value."""
            super().__init__(toml_file=toml_file, key=key)
            self.old: TomlValue | None = old
            self.new: TomlValue = new

    class Fail(TomlEvent):
        """General Failure."""

        __slots__ = ('failure',)

        def __init__(self, toml_file: TomlFile, failure: str) -> None:
            """Create a new :py:class:`TomlEvents.Fail` event with the given file and failure value."""
            super().__init__(toml_file=toml_file)
            self.failure = failure


class PathTomlDecoder(toml.TomlPreserveCommentDecoder):
    """Inherits the effects of :py:class:`toml.TomlPreserveCommentEncoder`.

    With native support for pathlib :py:class:`Path` values; not abandoning the TOML specification.
    """

    def load_value(self, v: str, strictly_valid: bool = True) -> tuple[Any, str]:
        """Load the given string value into its decoded type.

        If the value is a string and starts with the SPECIAL_PATH_PREFIX,
        load the value enclosed in quotes as a :py:class:`Path`.
        """
        if v[1:].startswith(_SPECIAL_PATH_PREFIX):
            v_path = Path(v[1:].removeprefix(_SPECIAL_PATH_PREFIX)[:-1])
            return v_path, 'path'
        return super().load_value(v=v, strictly_valid=strictly_valid)


class PathTomlEncoder(toml.TomlEncoder):
    """Combines both the effects of :py:class:`toml.TomlPreserveCommentEncoder` and :py:class:`toml.TomlPathlibEncoder`.

    Has native support for pathlib :py:class:`PurePath`; not abandoning the TOML specification.
    """

    def __init__(self, _dict=dict, preserve=False) -> None:
        """Map extra ``dump_funcs`` for :py:class:`CommentValue` and :py:class:`PurePath`."""
        super().__init__(_dict, preserve)
        self.dump_funcs[_CommentValue] = lambda comment_val: comment_val.dump(self.dump_value)
        self.dump_funcs[CommentValue] = lambda comment_val: comment_val.dump(self.dump_value)
        self.dump_funcs[PurePath] = self._dump_pathlib_path

    @staticmethod
    def _dump_pathlib_path(v: PurePath) -> str:
        """Translate :py:class:`PurePath` to string and dump."""
        # noinspection PyProtectedMember
        return toml.encoder._dump_str(str(v))  # type: ignore

    def dump_value(self, v: TomlValue) -> str:
        """Support :py:class:`Path` decoding by prefixing a :py:class:`PurePath` string with a special marker."""
        if isinstance(v, PurePath):
            if isinstance(v, Path):
                v = v.resolve()
            v = f'{_SPECIAL_PATH_PREFIX}{v}'
        return super().dump_value(v=v)


class TomlFile:
    """Object that manages the getting and setting of TOML configurations.

    Houses an :py:class:`EventBus` that allows you to subscribe Callables to changes in configuration.
    """

    def __init__(self, path: Path | str, default: dict[str, TomlValue | _CommentValue] | None = None) -> None:
        """Initialize a :py:class:`TomlFile` object.

        :param path: Path to the TOML file.
        :param default: Default values for the TOML file.
        """
        self._path: Path = Path(path)
        # FIXME: Default not working as expected during import
        self._data: dict[str, TomlValue | _CommentValue] = {} if default is None else default
        self.event_bus: EventBus[TomlEvents.TomlEvent] = EventBus()
        if not self.reload():
            warnings.warn(f'Could not load TOML file {self.path} on initialization.')

    def __getitem__(self, key: str) -> TomlValue:
        """Get the associated TOML value from the key."""
        return self.get(key)

    def __setitem__(self, key: str, value: TomlValue) -> None:
        """Set the key's associated TOML value."""
        self.set(key, value)

    def __delitem__(self, key: str) -> None:
        """Delete the key and it's associated TOML value."""
        del self._data[key]

    def _search_scope(self, path: str, mode: str) -> tuple[dict[str, TomlValue | _CommentValue], str]:
        """Search data for the given path to the value, and if found, return the scope and the key that path belongs to.

        :param path: Path to search data for.
        :param mode: Mode that determines which exceptions to raise.
        :return: Tuple containing the scope where the value is found, and the value key to access.
        :raises KeyError: If mode is 'set' and path is a table OR if mode is 'get' and path doesn't exist.
        :raises ValueError: If path is an empty string.
        """
        if not path:
            raise ValueError('Path cannot be an empty string.')

        key: str = path
        scope: dict[str, TomlValue | _CommentValue] = self._data
        paths: list[str] = path.split('/')

        if len(paths) > 1:
            for i, key in enumerate(paths):
                if key:
                    val: TomlValue | _CommentValue | None = scope.get(key)
                    if i == len(paths) - 1:
                        if mode == 'set' and isinstance(scope.get(key), dict):
                            raise KeyError(f'Cannot reassign Table "{".".join(paths[:i])}" to variable.')
                        if mode == 'get' and key not in scope:
                            raise KeyError(f'Key "{key}" not in Table "{".".join(paths[:i]) or "/"}".')

                    elif isinstance(val, dict):
                        scope = val  # type: ignore
                        continue

        return scope, key

    @property
    def path(self) -> Path:
        """:return: Current OS path that TomlFile with save and reload from."""
        return self._path

    @path.setter
    def path(self, value: Path | str) -> None:
        """Set the path to the TOML file to save and reload from.

        Translates string paths to pathlib Paths.
        """
        self._path = Path(value)

    def get(self, key: str) -> TomlValue:
        """Get a value from the key path. Searches with each '/' defining a new table to check.

        :param key: Key to get value from.
        :return: Value of key.
        :raises KeyError: If key doesn't exist.
        :raises ValueError: If key is an empty string.
        """
        scope, path = self._search_scope(key, mode='get')
        val: TomlValue | _CommentValue = scope[path]

        # Get value from _CommentValue
        if isinstance(val, _CommentValue):
            val = CommentValue.from_comment_val(val).val

        self.event_bus << TomlEvents.Get(self, key, val)

        return val

    def set(self, key: str, value: TomlValue, comment: str | None = None) -> None:
        """Set a key at path. Searches with each '/' defining a new table to check.

        :param key: Key to set.
        :param value: Value to set key as.
        :param comment: Append an optional comment to the value.
        :raises KeyError: If key evaluates to a table.
        :raises ValueError: If key is an empty string.
        """
        scope, path = self._search_scope(key, 'set')
        prev_val: TomlValue | _CommentValue | None = scope.get(path)
        new_val: TomlValue | _CommentValue = value

        if comment is not None:
            new_val = CommentValue(value, comment)

        # Preserve comments, or edit them if comment argument was filled
        elif isinstance(prev_val, _CommentValue):
            if comment is None:
                comment = prev_val.comment.lstrip().removeprefix(_COMMENT_PREFIX)

            new_val = CommentValue(value, comment, new_line=prev_val.comment.startswith('\n'))

        scope[path] = new_val

        self.event_bus << TomlEvents.Set(
            self, key,
            old=prev_val.val if isinstance(prev_val, _CommentValue) else prev_val,
            new=value
        )

    def save(self) -> bool:
        """Save current settings to self.path.

        :return: True if successful, otherwise False.
        """
        return self.export_to(self.path)

    def reload(self) -> bool:
        """Reset settings to settings stored in self.path.

        :return: True if successful, otherwise False.
        """
        return self.import_from(self.path, update=True)

    def export_to(self, path: Path | str) -> bool:
        """Export internal dictionary as a TOML file to path.

        :param path: Path to export TOML file to.
        :return: True if successful, otherwise False.
        """
        path = Path(path)  # Make sure path is of type Path
        if path.parent.is_dir():
            with path.open(mode='w', encoding='utf8') as file:
                toml.dump(self._data, file, encoder=PathTomlEncoder())

            self.event_bus << TomlEvents.Export(self, path)
            return True

        self.event_bus << TomlEvents.Fail(self, 'export')
        return False

    def import_from(self, path: Path | str, update: bool = False) -> bool:
        """Import TOML file from path to internal dictionary.

        :param path: Path to import TOML file from.
        :param update: If True, will update existing keys with new values, instead of replacing the internal dictionary.
        :return: True if successful, otherwise False.
        """
        path = Path(path)  # Make sure path is of type Path
        if path.is_file():
            try:
                with path.open(mode='r', encoding='utf8') as file:
                    toml_data = toml.load(file, decoder=PathTomlDecoder())
                    if update:
                        self._data |= toml_data
                    else:
                        self._data = toml_data

            except (LookupError, OSError, toml.TomlDecodeError):
                pass  # Pass to end of function, to fail.

            else:
                self.event_bus.fire(TomlEvents.Import(self, path))
                return True

        # If failed:
        self.event_bus << TomlEvents.Fail(self, 'import')
        return False



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""LicenseViewer implementation."""
from __future__ import annotations

__all__ = (
    'LicenseViewer',
)

from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *

from ...constants import *
from ...utils import init_layouts
from ...utils import init_objects
from ...utils import scroll_to_top
from ...utils import current_requirement_licenses
from ..aliases import app
from ..aliases import tr
from ..widgets import ExternalTextBrowser


# noinspection PyTypeChecker
class LicenseViewer(QWidget):
    """Widget that formats and shows the project's (and all of its requirements') license files."""

    def __init__(self, *args, **kwargs) -> None:
        """Create a new LicenseViewer."""
        super().__init__(*args, **kwargs)
        self.setWindowTitle(tr('gui.license_viewer.title'))
        self.setWindowIcon(app().icon_store['copyright'])
        self.resize(QSize(750, 800))

        self.license_data: list[tuple[str, str, str]] = []
        for pkg, licenses in current_requirement_licenses(HI_PACKAGE_NAME, include_extras=True).items():
            for title, text in licenses:
                self.license_data.append((pkg, title, text))

        self._init_ui()

    def _init_ui(self) -> None:
        self.license_dropdown: QComboBox = QComboBox(self)
        self.license_index_label: QLabel = QLabel(self)
        self.license_text_edit: ExternalTextBrowser = ExternalTextBrowser(self)

        init_objects({
            self.license_dropdown: {
                'size': {'fixed': (300, 26)},
                'activated': lambda i: self.view_index(i),
                'items': (
                    f'{pkg[0]}/{pkg[1]}' for pkg in self.license_data
                )
            },

            self.license_index_label: {
                'size': {'maximum': (50, None)},
                'text': tr(
                    'gui.license_viewer.current_index', self.license_dropdown.currentIndex() + 1, len(self.license_data)
                )
            },

            self.license_text_edit: {
                'font': QFont('consolas', 11),
                'openExternalLinks': True,
            },

            (next_license_button := QPushButton(self)): {
                'size': {'fixed': (100, None)},
                'clicked': self.next_license,
            },

            (prev_license_button := QPushButton(self)): {
                'size': {'fixed': (100, None)},
                'clicked': self.prev_license,
            }
        })

        app().init_translations({
            self.setWindowTitle: 'gui.license_viewer.title',

            next_license_button.setText: 'gui.license_viewer.next',
            prev_license_button.setText: 'gui.license_viewer.previous',
        })

        init_layouts({
            # Top widget layouts
            (top_right := QHBoxLayout()): {
                'items': [prev_license_button, next_license_button, self.license_index_label]
            },
            (top := QHBoxLayout()): {
                'items': [self.license_dropdown, top_right]
            },

            # Main layout
            QVBoxLayout(self): {
                'items': [top, self.license_text_edit]
            }
        })

        self.license_text_edit.connect_key_to(Qt.Key.Key_Left, self.prev_license)
        self.license_text_edit.connect_key_to(Qt.Key.Key_Right, self.next_license)
        self.view_current_index()

        top_right.setAlignment(Qt.AlignmentFlag.AlignRight)

        cursor = self.license_text_edit.textCursor()
        cursor.clearSelection()
        cursor.select(QTextCursor.SelectionType.Document)

    def next_license(self) -> None:
        """View the next license."""
        index = self.license_dropdown.currentIndex() + 1
        if index + 1 > self.license_dropdown.count():
            index = 0
        self.view_index(index)

    def prev_license(self) -> None:
        """View the previous license."""
        if (index := self.license_dropdown.currentIndex() - 1) < 0:
            index = self.license_dropdown.count() - 1
        self.view_index(index)

    def view_index(self, index: int) -> None:
        """View the license at the given index."""
        self.license_dropdown.setCurrentIndex(index)
        self.view_current_index()

    def view_current_index(self) -> None:
        """View the license data at the current index."""
        current_data: tuple[str, str, str] = self.license_data[self.license_dropdown.currentIndex()]

        license_text = current_data[2] or tr('gui.license_viewer.not_found')
        self.license_index_label.setText(tr(
            'gui.license_viewer.current_index', self.license_dropdown.currentIndex() + 1, len(self.license_data)
        ))

        # Use Regex & HTML to make links clickable
        output = license_text
        replaced = set()
        for match in HI_URL_PATTERN.finditer(license_text):
            if (match := match[0]) not in replaced:
                output = output.replace(match, f'<a href="{match}" style="color: #2A5DB0">{match}</a>')
                replaced.add(match)

        stripped_output = ''
        for line in output.splitlines():
            stripped_output += line.strip() + '\n'

        stripped_output = stripped_output.strip()
        self.license_text_edit.setHtml(
            f'<body style="white-space: pre-wrap">'
            f'<center>{stripped_output}</center>'
            f'</body>'
        )

        scroll_to_top(self.license_text_edit)



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""An independent script to upgrade the version of a given package, then launch it."""

import subprocess
import sys
from pathlib import Path


def main(package_to_install: str, module_to_run: str | None = None) -> None:
    """Script entry point. if __name__ == '__main__'."""
    python = Path(sys.executable)

    # Ensure that pip is ran under windowless, so it doesn't install GUI scripts directing to "pythonww.exe".
    # This is due to pip appending a "w" to its own executable path.
    pip_python = python.with_stem(python.stem.removesuffix('w'))

    subprocess.run(  # nosec B603:subprocess_without_shell_equals_true
        (pip_python, '-m', 'pip', 'install', '--upgrade', package_to_install),
        capture_output=True, text=True, check=True
    )

    if module_to_run is not None:
        subprocess.run(  # nosec B603:subprocess_without_shell_equals_true
            (python, '-m', module_to_run),
            check=False
        )


if __name__ == '__main__':
    main(*sys.argv[1:3])



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""ReadmeViewer implementation."""
from __future__ import annotations

__all__ = (
    'ReadmeViewer',
)

from importlib.metadata import metadata
from typing import Final

from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *

from ...constants import *
from ...utils import init_layouts
from ...utils import init_objects
from ..aliases import app
from ..aliases import tr
from ..widgets import ExternalTextBrowser


class ReadmeViewer(QWidget):
    """Widget that formats and shows the project's README.md, stored in the projects 'Description' metadata tag."""

    README_TEXT: Final[str] = metadata(HI_PACKAGE_NAME)['Description']

    def __init__(self, *args, **kwargs) -> None:
        """Create a new :py:class:`ReadmeViewer` and initialize UI."""
        super().__init__(*args, **kwargs)
        self.setWindowTitle(tr('gui.readme_viewer.title'))
        self.setWindowIcon(app().get_theme_icon('message_information') or
                           self.style().standardIcon(QStyle.StandardPixmap.SP_DialogApplyButton))
        self.resize(QSize(750, 750))
        self._init_ui()

    def _init_ui(self) -> None:
        self.readme_viewer = ExternalTextBrowser(self)

        init_objects({
            (close_button := QPushButton(self)): {
                'size': {'minimum': (None, 40)},
                'font': QFont(close_button.font().family(), 16),
                'clicked': self.close
            },

            self.readme_viewer: {
                'font': QFont(self.readme_viewer.font().family(), 10),
                'openExternalLinks': True
            }
        })

        app().init_translations({
            self.setWindowTitle: 'gui.readme_viewer.title',

            close_button.setText: 'gui.readme_viewer.close'
        })

        init_layouts({
            # Main layout
            QVBoxLayout(self): {
                'items': [self.readme_viewer, close_button]
            }
        })

        self.readme_viewer.set_hot_reloadable_text(self.README_TEXT, 'markdown')



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""Settings window implementation."""
from __future__ import annotations

__all__ = (
    'SettingsWindow',
)

from pathlib import Path

from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *

from ...constants import *
from ...events import EventBus
from ...models import DeferredCallable
from ...models import Singleton
from ...tomlfile import TomlEvents
from ...utils import init_layouts
from ...utils import init_objects
from ..aliases import app
from ..aliases import tr
from ..widgets import PasteLineEdit
from ..widgets import TranslatableComboBox


# noinspection PyTypeChecker
class SettingsWindow(Singleton, QWidget):
    """Window that provides user interaction with the application's settings.

    :py:class:`SettingsWindow` is a singleton and can be accessed via the class using the
    SettingsWindow.instance() class method.
    """

    _singleton_base_type = QWidget
    _singleton_check_ref = False

    def __init__(self, size: QSize) -> None:
        """Create the settings window."""
        super().__init__()

        self.setWindowTitle(tr('gui.settings.title'))
        self.setWindowIcon(app().icon_store['settings'])
        self.resize(size)
        self.setFixedWidth(self.width())

        # Show an error dialog on import failure
        EventBus['settings'].subscribe(
            DeferredCallable(app().show_dialog, 'errors.settings.import_failure', self,
                             description_args=(app().settings.path,)),
            TomlEvents.Fail, event_predicate=lambda event: event.failure == 'import'
        )

        self._init_ui()

    def _init_ui(self) -> None:

        def import_settings() -> None:
            """Import settings from a chosen TOML file."""
            file_path = Path(QFileDialog.getOpenFileName(self, caption=tr('gui.settings.import'),
                                                         dir=str(HI_CONFIG_PATH),
                                                         filter='TOML Files (*.toml);;All files (*.*)')[0])
            if file_path.is_file():
                if app().settings.import_from(file_path):
                    save_button.setDisabled(False)

        def export_settings() -> None:
            """Export current settings to a chosen file location."""
            file_path = Path(QFileDialog.getSaveFileName(self, caption=tr('gui.settings.export'),
                                                         dir=str(HI_CONFIG_PATH),
                                                         filter='TOML Files (*.toml);;All files (*.*)')[0])
            if str(file_path) != '.':
                app().settings.export_to(file_path)

        def hide_key() -> None:
            """Hide API key."""
            self.key_set_button.setDisabled(True)
            self.key_field.setDisabled(True)
            self.key_field.setText(app().client.hidden_key())
            self.key_field.setAlignment(Qt.AlignmentFlag.AlignCenter)

        def toggle_key_visibility() -> None:
            """Toggle hiding and showing the API key."""
            if not self.key_field.isEnabled():
                self.key_field.setAlignment(Qt.AlignmentFlag.AlignLeft)
                self.key_field.setText(str(app().client.wpauth))
                self.key_field.setDisabled(False)
                self.key_field.setFocus()
                self.key_set_button.setDisabled(False)
            else:
                hide_key()

        def set_key() -> None:
            """Set the client's auth_key to the current text in the key field."""
            if text := self.key_field.text().strip():
                app().client.wpauth = text
            else:
                del app().client.wpauth
            toggle_key_visibility()

        def clear_token() -> None:
            del app().client.token
            self.token_clear_button.setDisabled(True)

        # Define widget attributes
        # Cannot be defined in init_objects() as walrus operators are not allowed for object attribute assignment.
        # This works in the standard AST, but is a seemingly arbitrary limitation set by the interpreter.
        # See: https://stackoverflow.com/questions/64055314#answer-66617839
        (
            self.key_set_button, self.token_clear_button,
            self.theme_dropdown, self.aspect_ratio_dropdown, self.transformation_dropdown,
            self.line_wrap_dropdown, self.icon_mode_dropdown,
            self.key_field
        ) = (
            QPushButton(self), QPushButton(self),
            TranslatableComboBox(self), TranslatableComboBox(self), TranslatableComboBox(self),
            TranslatableComboBox(self), TranslatableComboBox(self),
            PasteLineEdit(self)
        )

        init_objects({
            # Labels
            (theme_label := QLabel(self)): {
                'size': {'maximum': (85, None)}
            },
            (aspect_ratio_label := QLabel(self)): {
                'size': {'maximum': (90, None)}
            },
            (transformation_label := QLabel(self)): {
                'size': {'maximum': (90, None)}
            },
            (line_wrap_label := QLabel(self)): {
                'size': {'maximum': (90, None)}
            },
            (icon_mode_label := QLabel(self)): {
                'size': {'maximum': (90, None)}
            },

            # Buttons
            (save_button := QPushButton(self)): {
                'disabled': True,
                'size': {'maximum': (50, None)},
                'clicked': app().settings.save
            },
            (reload_button := QPushButton(self)): {
                'size': {'maximum': (60, None)},
                'clicked': app().settings.reload
            },
            (import_button := QPushButton(self)): {
                'clicked': import_settings
            },
            (export_button := QPushButton(self)): {
                'clicked': export_settings
            },
            (open_editor_button := QPushButton(self)): {
                'clicked': DeferredCallable(QDesktopServices.openUrl, lambda: QUrl(app().settings.path.as_uri()))
            },
            (key_show_button := QPushButton(self)): {
                'clicked': toggle_key_visibility
            },
            self.key_set_button: {
                'size': {'minimum': (40, None)},
                'clicked': set_key
            },
            self.token_clear_button: {
                'disabled': not app().client.token,
                'clicked': clear_token
            },

            # Line editors
            self.key_field: {
                'font': QFont('segoe ui', 8), 'text': app().client.hidden_key(),
                'pasted': set_key, 'returnPressed': self.key_set_button.click,
                'size': {'minimum': (220, None)}, 'alignment': Qt.AlignmentFlag.AlignCenter
            },

            # Dropdowns
            self.theme_dropdown: {
                'activated': DeferredCallable(
                    app().settings.__setitem__,
                    'gui/themes/selected',
                    lambda: app().sorted_themes()[self.theme_dropdown.currentIndex()].id
                ),
                'items': (theme.display_name for theme in app().sorted_themes())
            },
            self.aspect_ratio_dropdown: {
                'activated': DeferredCallable(
                    app().settings.__setitem__,
                    'gui/media_output/aspect_ratio_mode',
                    self.aspect_ratio_dropdown.currentIndex
                ),
                'items': (
                    'gui.settings.media.aspect_ratio.ignore',
                    'gui.settings.media.aspect_ratio.keep',
                    'gui.settings.media.aspect_ratio.expanding'
                )
            },
            self.transformation_dropdown: {
                'activated': DeferredCallable(
                    app().settings.__setitem__,
                    'gui/media_output/transformation_mode',
                    self.transformation_dropdown.currentIndex
                ),
                'items': (
                    'gui.settings.media.image_transform.fast',
                    'gui.settings.media.image_transform.smooth'
                )
            },
            self.line_wrap_dropdown: {
                'activated': DeferredCallable(
                    app().settings.__setitem__,
                    'gui/text_output/line_wrap_mode',
                    self.line_wrap_dropdown.currentIndex
                ),
                'items': (
                    'gui.settings.text.line_wrap.no_wrap',
                    'gui.settings.text.line_wrap.widget',
                    'gui.settings.text.line_wrap.fixed_pixel',
                    'gui.settings.text.line_wrap.fixed_column'
                )
            },
            self.icon_mode_dropdown: {
                'activated': DeferredCallable(
                    app().settings.__setitem__,
                    'gui/cache_explorer/icon_mode',
                    self.icon_mode_dropdown.currentIndex
                ),
                'items': (
                    'gui.settings.cache.icon_mode.no_icons',
                    'gui.settings.cache.icon_mode.default',
                    'gui.settings.cache.icon_mode.preview'
                )
            }
        })

        app().init_translations({
            self.setWindowTitle: 'gui.settings.title',
            self.aspect_ratio_dropdown.translate_items: '',
            self.line_wrap_dropdown.translate_items: '',
            self.theme_dropdown.translate_items: '',
            self.transformation_dropdown.translate_items: '',

            # Labels
            theme_label.setText: 'gui.settings.theme',
            aspect_ratio_label.setText: 'gui.settings.media.aspect_ratio',
            transformation_label.setText: 'gui.settings.media.image_transform',
            line_wrap_label.setText: 'gui.settings.text.line_wrap',
            icon_mode_label.setText: 'gui.settings.cache.icon_mode',

            # Buttons
            save_button.setText: 'gui.settings.save',
            reload_button.setText: 'gui.settings.reload',
            import_button.setText: 'gui.settings.import',
            export_button.setText: 'gui.settings.export',
            open_editor_button.setText: 'gui.settings.open_editor',
            key_show_button.setText: 'gui.settings.auth.edit',
            self.key_set_button.setText: 'gui.settings.auth.set',
            self.token_clear_button.setText: 'gui.settings.auth.clear_token'
        })

        init_layouts({
            # Add bottom widgets
            (token_layout := QHBoxLayout()): {
                'items': [self.token_clear_button]
            },
            (key_layout := QHBoxLayout()): {
                'items': [self.key_field, self.key_set_button]
            },
            (bottom := QVBoxLayout()): {
                'items': [key_show_button, key_layout, token_layout]
            },

            # Add middle widgets
            (cache_layout := QHBoxLayout()): {
                'items': [icon_mode_label, self.icon_mode_dropdown]
            },
            (output_layout := QGridLayout()): {
                'items': [
                    (aspect_ratio_label, 0, 0),
                    (self.aspect_ratio_dropdown, 0, 10),
                    (transformation_label, 0, 20),
                    (self.transformation_dropdown, 0, 30),
                    (line_wrap_label, 10, 0),
                    (self.line_wrap_dropdown, 10, 10)
                ]
            },
            (theme_layout := QHBoxLayout()): {
                'items': [theme_label, self.theme_dropdown]
            },
            (middle := QVBoxLayout()): {
                'items': [theme_layout, output_layout, cache_layout]
            },

            # Add top widgets
            (io_buttons := QHBoxLayout()): {
                'items': [save_button, reload_button, import_button, export_button]
            },
            (top := QVBoxLayout()): {
                'items': [io_buttons, open_editor_button]
            },

            # Main layout
            QGridLayout(self): {
                'items': [
                    (top, 0, 0, Qt.AlignmentFlag.AlignTop),
                    (middle, 10, 0, Qt.AlignmentFlag.AlignTop),
                    (bottom, 20, 0, Qt.AlignmentFlag.AlignBottom)
                ]
            }
        })

        EventBus['settings'].subscribe(
            DeferredCallable(save_button.setDisabled, False),
            TomlEvents.Set, lambda event: event.old != event.new)

        EventBus['settings'].subscribe(
            DeferredCallable(save_button.setDisabled, True),
            TomlEvents.Export, lambda event: event.toml_file.path == event.path)

        EventBus['settings'].subscribe(
            DeferredCallable(save_button.setDisabled, True),
            TomlEvents.Import, lambda event: event.toml_file.path == event.path)

        EventBus['settings'].subscribe(
            DeferredCallable(self.refresh_dropdowns),
            TomlEvents.Import)

    def refresh_dropdowns(self) -> None:
        """Refresh all dropdown widgets with the current settings assigned to them."""
        settings = app().settings
        self.icon_mode_dropdown.setCurrentIndex(settings['gui/cache_explorer/icon_mode'])               # type: ignore
        self.aspect_ratio_dropdown.setCurrentIndex(settings['gui/media_output/aspect_ratio_mode'])      # type: ignore
        self.transformation_dropdown.setCurrentIndex(settings['gui/media_output/transformation_mode'])  # type: ignore
        self.line_wrap_dropdown.setCurrentIndex(settings['gui/text_output/line_wrap_mode'])             # type: ignore
        self.theme_dropdown.setCurrentIndex(app().theme_index_map[settings['gui/themes/selected']])     # type: ignore

    # # # # # Events

    def showEvent(self, event: QShowEvent) -> None:
        """Auto hides the key upon un-minimizing."""
        super().showEvent(event)
        self.refresh_dropdowns()

        init_objects({
            self.key_field: {
                'disabled': True,
                'alignment': Qt.AlignmentFlag.AlignCenter,
                'text': app().client.hidden_key()
            },
            self.key_set_button: {'disabled': True},
            self.token_clear_button: {'disabled': not app().client.token}
        })

        event.accept()



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""Module containing code related to translation and language files."""
from __future__ import annotations

__all__ = (
    'default_language_file',
    'format_value',
    'Language',
    'to_lang',
    'Translator',
)

import json
import re
from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path
from typing import Any
from typing import Final

from .constants import *

_POS_PARAM_REF: Final[re.Pattern] = re.compile(r'{([1-9]\d*|0)}')
"""Pattern for finding positional parameter references."""

_KEY_REF: Final[re.Pattern] = re.compile(r'{[\w\-.]*}')
"""Pattern for finding lang key references."""

_LANG_PATH: Final[Path] = HI_RESOURCE_PATH / 'lang'
"""Directory containing language JSON data."""


def default_language_file() -> Path:
    """Return the first path found that has ".default" in its extension.

    ex: en.default.json

    :raises FileNotFoundError: If .default file is not found.
    """
    default_lang_file: Path | None = None
    for lang_file in _LANG_PATH.iterdir():
        if '.default' in lang_file.suffixes:
            default_lang_file = lang_file
            break

    if default_lang_file is None:
        raise FileNotFoundError('No file with the suffix of .default.json found.')

    return default_lang_file


def format_value(
        value: str,
        *args: Any,
        _language: Language | None = None,
        _key_eval: bool = True
) -> str:
    r"""Format a str with positional arguments.

    You can use {0} notation to refer to a specific positional argument.
    If the notation chars are replaced with the argument, you can no longer use it as a normal positional argument.

    JSON strings ex::

        {
            "a": "{0} is the same as {0} using only one argument.",
            "b": "{0} %s will not work and require 2 arguments",
            "c": "%s {0} is the same as above, where \"%s\" is now the 2nd argument since the 1st is used by \"{0}\""
        }
    """
    list_args: list[Any] = list(args)
    replaced: set[str] = set()

    for matches in _POS_PARAM_REF.finditer(value):
        match: str = matches[0]
        arg_val: Any = args[int(match.strip('{}'))]
        if match not in replaced:
            replaced.add(match)
            value = value.replace(match, arg_val)
            list_args.remove(arg_val)

    value = value % tuple(list_args)

    if _key_eval and _language is not None:
        replaced.clear()
        for _ in range(50):
            # Loop if a match is found, to recursively get translation key values.
            # Limit of 50 to prevent an infinite loop.

            found: bool = False
            for rec_matches in _KEY_REF.finditer(value):
                found = True
                if (rec_match := rec_matches[0]) not in replaced:
                    replaced.add(rec_match)
                    value = value.replace(rec_match, _language.get_raw(rec_match.strip('{}')))

            if not found:
                # End loop if no inner translation keys are found.
                break

    return value


def matched_subtags(tag: str) -> dict[str, str]:
    """Match all subtags in a valid language tag and returns a dictionary representation.

    :param tag: Language tag which follows RFC 5646.
    :return: Dictionary containing subtags and their values.
    :raises ValueError: If ``tag`` is not a valid language tag.
    """
    if (match := RFC_5646_PATTERN.match(tag)) is not None:
        return match.groupdict()
    raise ValueError(f'"{tag}" is not a valid language tag (RFC 5646).')


def to_lang(language: str | Language) -> Language:
    """Assert that a given value is a :py:class:`Language`."""
    if not isinstance(language, Language):
        language = str(language)  # Stringify non-language object

    if isinstance(language, str):
        language = Language.from_tag(language)
    return language


class Language:
    """Object containing language data. Retrieves stored data with associated RFC 5646 language tags."""

    def __init__(
            self,
            primary: str,
            region: str | None = None,
            _selected_file: Path | None = None,
            **kwargs: str | None
    ) -> None:
        """Build a new Language object using given sub-tags.

        For more information on RFC 5646, visit https://datatracker.ietf.org/doc/html/rfc5646

        :param primary: The first part of the language code. i.e. [en]-US | Section 2.2.1
        :param region: The (commonly) second part of the language code. i.e. en-[US] | Section 2.2.4

        :keyword extlang: Extension to primary subtag | Section 2.2.2
        :keyword script: Script used to write the language | Section 2.2.3
        :keyword variants: A variant of the given language subtag | Section 2.2.5
        :keyword extensions: Extension to the given language, prefixed with a singleton | Section 2.2.6
        :keyword private: Any subtag that is not publicly defined | Section 2.2.7
        """
        subtags: dict[str, str | None] = {'primary': primary, 'region': region}
        subtags.update(kwargs)

        self._data: dict[str, str] = {}
        self.tag: str = Language.build_tag(subtags)
        self.subtags: dict[str, str] = matched_subtags(self.tag)

        if _selected_file is None:
            _selected_file = self.closest_file()

        if _selected_file is not None:
            # Read the language file corresponding to this Language's tags.
            file_data: dict[str, Any] = json.loads(_selected_file.read_text(encoding='utf8'))

            # Read the parent language file if it exists.
            if (parent_lang := file_data['meta'].get('inherits_from')) is not None:
                self._data |= Language.from_tag(parent_lang).data

            # Overwrite inherited keys.
            self._data |= file_data['keys']

    def __repr__(self) -> str:
        """Representation of the :py:class:`Language` with its tag and number of loaded language keys."""
        return f'<Language data for {self.tag} ({len(self.data)} keys)>'

    def __getitem__(self, key: str) -> str:
        """Get the raw value of the given key (No formatting)."""
        return self.get_raw(key)

    @classmethod
    def from_tag(cls, tag: str) -> Language:
        """Build a :py:class:`Language` object using a plain string tag.

        Breaks tag into sub-tags and verifies compliance with RFC 5646.

        :raises ValueError: If tag is not a valid language tag.
        """
        return cls(**matched_subtags(tag))

    @classmethod
    def from_path(cls, path: Path | str) -> Language:
        """Build a :py:class:`Language` object using a plain string tag.

        Breaks tag into sub-tags and verifies compliance with RFC 5646.

        :raises ValueError: If tag from the path's filename is not a valid language tag.
        """
        path = Path(path)
        tag: str = Path(path).stem.split('.')[0].replace('_', '-')
        return cls(**matched_subtags(tag), _selected_file=path)

    @staticmethod
    def build_tag(tag_dict: dict[str, str | None]) -> str:
        """Build tag from a tag dictionary.

        :param tag_dict: Dictionary obtained from RFC_5646_PATTERN.match(tag).groupdict()
        :return: String representation of tag with correct case formatting.

        :raises ValueError: If a given subtag is invalid.
        """
        (primary, extlang, script, region, variants, extensions, private) = (
            tag_dict.pop('primary', None),
            tag_dict.pop('extlang', None),
            tag_dict.pop('script', None),
            tag_dict.pop('region', None),
            tag_dict.pop('variants', None),
            tag_dict.pop('extensions', None),
            tag_dict.pop('private', None),
        )

        checked: set[str] = set()
        subtags: list[str] = []
        err: Exception | None = None

        if primary is None and private is None:
            err = ValueError('The primary and/or private subtag must be filled out.')

        if primary is not None:
            subtags.append(primary.lower())

        if extlang is not None:
            subtags.append(extlang.lower())

        if script is not None:
            subtags.append(script.title())

        if region is not None:
            subtags.append(region.upper())

        if variants is not None:
            for subtag in variants.split('-'):
                # RFC 5646 section 2.2.5.5
                if (variant := subtag.lower()) in checked:
                    err = ValueError(f'Variant subtag "{subtag}" is repeated.')

                checked.add(variant)
                subtags.append(variant)
            checked.clear()

        if extensions is not None:
            for i, subtag in enumerate(extensions.split('-')):
                if not i % 2:
                    # RFC 5646 section 2.2.6.3
                    if (singleton := subtag.lower()) in checked:
                        err = ValueError(f'Singleton subtag "{subtag}" is repeated.')

                    checked.add(singleton)
                subtags.append(subtag.lower())
            checked.clear()

        if private is not None:
            for subtag in private.split('-'):
                subtags.append(subtag.lower())

        tag: str = '-'.join(subtags)

        if err is not None:
            raise ValueError(str(err)[:-1] + f' in language tag "{tag}".') from err

        return tag

    @property
    def data(self) -> dict[str, Any]:
        """Return a copy of the internal key dictionary."""
        return self._data.copy()

    def closest_file(self) -> Path | None:
        """Find the most relevant file in the ``_LANG_DIR`` to represent this :py:class:`Language`'s tag.

        The primary language subtag MUST match for a lang file to be considered. ex::

            "en-Latn-US-1994" would regress to en_us.json, as they both share their primary and region subtags.
            "es-US" would NOT regress to en_us.json, as they do not share their primary subtags.

        """
        most_relevant: tuple[Path | None, int] = (None, 0)

        for lang_file in _LANG_PATH.iterdir():
            if lang_file.suffix == '.json':
                file_tag: str = lang_file.stem.split('.')[0].replace('_', '-').lower()

                # For exact match
                if self.tag.lower() == file_tag:
                    most_relevant = (lang_file, 1000)
                    break

                file_subtags: dict[str, str] = matched_subtags(file_tag)

                # Count how close our tag is to this file's tag
                intersection = {a: b for ((a, b), (c, d))
                                in zip(self.subtags.items(), file_subtags.items())
                                if a == c and b is not None and d is not None and b.lower() == d.lower()}

                # Ensure that the primary language subtag is the same
                if intersection.get('primary') is not None:
                    relevance = len(intersection)
                else:
                    relevance = 0

                if relevance > most_relevant[1]:
                    most_relevant = (lang_file, relevance)

        return most_relevant[0]

    def get(self, key: str, *args: Any, default: str | None = None, key_eval: bool = True) -> str:
        """Get a translation key and format with the given arguments if required.

        :param key: Key to get in JSON data.
        :param args: Formatting arguments; packed into tuple and used with str modulus.
        :param default: Default string if key doesn't exist. (Is not formatted)
        :param key_eval: If False, don't recursively evaluate translation keys.
        """
        # Default value is key if not overridden
        default = default if default is not None else key

        if (result := self._data.get(key, default)) is not default:
            return format_value(result, *args, _language=self, _key_eval=key_eval)

        # Dont format default value
        quote = '"'
        for arg in (f'%{str(arg) if not isinstance(arg, str) else quote + arg + quote}%' for arg in args):
            result += arg
        return result

    def get_raw(self, key: str) -> str:
        """Get raw result, no formatting or defaults.

        :raises KeyError: If key is not in JSON data.
        """
        return self._data[key]


class Translator:
    """Simple class meant to abstract key translation into a function-like state.

    Usage::
        translate = Translator('en-US')
        translate('a.translation.key') -> american value
        translate.language = 'de-GER'
        translate('a.translation.key') -> german value
    """

    def __init__(self, language: Language | str | None = None) -> None:
        """Create a :py:class:`Translator` for the given :py:class:`Language`.

        If not langauge is supplied, it defaults to the default_language_file.

        :param language: Language to use for translator.
        """
        if language is not None:
            self._language = to_lang(language)
        else:
            self._language = Language.from_path(default_language_file())

    def __bool__(self) -> bool:
        """Return whether the :py:class:`Translator` is available."""
        return True

    def __call__(self, key: str, *args: Any, **kwargs) -> str:
        """Syntax sugar for get_translation."""
        return self.get_translation(key, *args, **kwargs)

    def __repr__(self) -> str:
        """Representation of the :py:class:`Translator` with its tag and number of loaded language keys."""
        language = self.language
        return f'<Translator {language=}>'

    @property
    def language(self) -> Language:
        """Return current :py:class:`Language` for this translator."""
        return self._language

    @language.setter
    def language(self, value: Language | str) -> None:
        self._language = to_lang(value)

    def get_translation(self, key: str, *args: Any, **kwargs) -> str:
        """Get a translation key's value for the current language."""
        return self.language.get(key, *args, **kwargs)

    @contextmanager
    def as_language(self, language: Language | str) -> Iterator[None]:
        """Temporarily translate for a specific language using a context manager."""
        # __enter__
        old_lang = self._language
        self.language = language
        yield

        # __exit__
        self._language = old_lang



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""ExceptionLogger implementation."""
from __future__ import annotations

__all__ = (
    'ExceptionLogger',
    'LoggedException',
)

from datetime import datetime
from types import TracebackType
from typing import NamedTuple

from PySide6.QtWidgets import *

from ...events import EventBus
from ...exception_hook import ExceptionEvent
from ..aliases import tr


class ExceptionLogger(QPushButton):
    """A :py:class:`QPushButton` that logs exceptions to the event bus."""

    level_icon_list: list = [
        QStyle.StandardPixmap.SP_MessageBoxInformation,  # 0, Not a concern
        QStyle.StandardPixmap.SP_MessageBoxWarning,      # 1, Warning
        QStyle.StandardPixmap.SP_MessageBoxCritical      # 2, Error
    ]

    def __init__(self, *args, **kwargs) -> None:
        """Initialize the ExceptionLogger."""
        from ..windows import ExceptionReporter

        super().__init__(*args, **kwargs)
        EventBus['exceptions'].subscribe(self.on_exception, ExceptionEvent)

        self.label: QLabel = QLabel(self)
        self.exception_log: list[LoggedException] = []
        self.reporter: ExceptionReporter = ExceptionReporter(self)
        self.severity = 0

        self.reporter.setMinimumWidth(300)

    def clear_exceptions(self) -> None:
        """Clear the exception log and disable the button."""
        self.exception_log.clear()
        self.severity = 0
        self.setText('')
        self.label.setText(tr('gui.status.default'))

    def remove_exception(self, index: int) -> None:
        """Remove an exception from the log and update the current severity."""
        if self.exception_log:
            self.exception_log.pop(index)

        if not self.exception_log:
            self.clear_exceptions()
        else:
            logged = len(self.exception_log)
            self.setText(f'({logged})' if logged < 10 else '(9+)')
            self.severity = self.exception_log[0].severity

    def on_exception(self, event: ExceptionEvent) -> None:
        """Update the exception log and change set the max level."""
        level = 1 if isinstance(event.exception, Warning) and self.severity < 1 else 2

        self.severity = max(self.severity, level)

        self.exception_log.append(LoggedException(level, event.exception, event.traceback, datetime.now()))
        self.sort_exceptions()

        logged = len(self.exception_log)
        self.setText(f'({logged})' if logged < 10 else '(9+)')

    def sort_exceptions(self) -> None:
        """Sort the exception log by severity."""
        if self.exception_log:
            self.exception_log = list(reversed(sorted(self.exception_log, key=lambda x: x.severity)))
            self.severity = self.exception_log[0].severity
        else:
            self.severity = 0

    @property
    def severity(self) -> int:
        """Get the max level of the exception log."""
        return self._severity

    @severity.setter
    def severity(self, value: int) -> None:
        """Set the max level of the exception log and update the icon."""
        self._severity = value
        self.setIcon(self.style().standardIcon(self.level_icon_list[self.severity]))
        self.reporter.setWindowIcon(self.icon())


class LoggedException(NamedTuple):
    """Container for a logged exception.

    Includes the severity of the exception, the exception itself, an optional traceback, and a timestamp.
    """

    severity: int = 0
    exception: BaseException | None = None
    traceback: TracebackType | None = None
    timestamp: datetime | None = None



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""Package containing :py:class:`QMenu` Context Menus."""

__all__ = (
    'CacheIndexContextMenu',
    'ColumnContextMenu',
    'FileContextMenu',
    'HelpContextMenu',
    'ToolsContextMenu',
)

from .cache_index import CacheIndexContextMenu
from .column import ColumnContextMenu
from .file import FileContextMenu
from .help import HelpContextMenu
from .tools import ToolsContextMenu



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""Miscellaneous objects for hi_getter."""
from __future__ import annotations

__all__ = (
    'CaseInsensitiveDict',
    'DeferredCallable',
    'DistributedCallable',
    'Singleton',
)

import sys
import weakref
from abc import ABC
from abc import abstractmethod
from collections import OrderedDict
from collections.abc import Callable
from collections.abc import Collection
from collections.abc import Generator
from collections.abc import Iterable
from collections.abc import Mapping
from collections.abc import MutableMapping
from typing import Any
from typing import Generic
from typing import NoReturn
from typing import TypeAlias
from typing import TypeVar
from warnings import warn

_VT = TypeVar('_VT')
_CT = TypeVar('_CT', bound=Collection[Callable])  # Bound to Collection of Callables
_PT = TypeVar('_PT')  # Positional Arguments
_KT = TypeVar('_KT')  # Keyword Arguments
_PTCallable: TypeAlias = Callable[[], _PT]  # Which returns _PT
_KTCallable: TypeAlias = Callable[[], _KT]  # Which returns _KT


class _AbstractCallable(ABC):
    """Abstract :py:class:`Callable` object.

    Child objects are callable as a shortcut to their `run` method.
    """

    def __call__(self, *args, **kwargs) -> Any:
        """Syntax sugar for self.run()."""
        return self.run(*args, **kwargs)

    @abstractmethod
    def run(self, *args, **kwargs) -> Any:
        """Execute functionality with the provided arguments and return the result.

        :param args: positional arguments to call with.
        :param kwargs: keyword arguments to call with.
        """
        raise NotImplementedError

##########
# NOTICE:
##########
# Requests
# Copyright 2019 Kenneth Reitz
# Apache 2.0 License
# https://github.com/psf/requests/blob/main/LICENSE


class CaseInsensitiveDict(MutableMapping, Generic[_VT]):
    """A case-insensitive :py:class:`dict`-like object.

    Implements all methods and operations of
    :py:class:`MutableMapping` as well as :py:class:`dict`'s ``copy``. Also
    provides ``lower_items``.

    All keys are expected to be strings. The structure remembers the
    case of the last key to be set, and ``iter(instance)``,
    ``keys()``, ``items()``, ``iterkeys()``, and ``iteritems()``
    will contain case-sensitive keys. However, querying and contains
    testing is case-insensitive::

        cid = CaseInsensitiveDict()
        cid['Accept'] = 'application/json'
        cid['aCCEPT'] == 'application/json'  # True
        list(cid) == ['Accept']  # True

    For example, ``headers['content-encoding']`` will return the
    value of a ``'Content-Encoding'`` response header, regardless
    of how the header name was originally stored.

    If the constructor, ``.update``, or equality comparison
    operations are given keys that have equal ``.lower()``s, the
    behavior is undefined.
    """

    __slots__ = ('_store',)

    def __init__(self,
                 data: Iterable[tuple[str, _VT]] | Mapping[str, _VT] | None = None,
                 **kwargs
                 ) -> None:
        """Create a :py:class:`CaseInsensitiveDict` with the given data.

        :param data: Data to turn into a CaseInsensitiveDict
        :param kwargs: key-values in form of kwargs
        """
        self._store = OrderedDict()
        if data is None:
            data = {}
        self.update(data, **kwargs)

    def __setitem__(self, key: str, value: _VT) -> None:
        """Set the associated value of the given key.

        Use the lowercased key for lookups, but store the actual key alongside the value.
        """
        self._store[key.lower()] = (key, value)

    def __getitem__(self, key: str) -> _VT:
        """Get the associated value of the given key.

        Use the lowercased key for lookups.
        """
        return self._store[key.lower()][1]

    def __delitem__(self, key: str) -> None:
        """Delete the key and associated value from the dictionary."""
        del self._store[key.lower()]

    def __iter__(self) -> Generator[str, None, None]:
        """Return a :py:class:`Generator` containing all original (cased) keys."""
        return (cased_key for cased_key, mapped_value in self._store.values())

    def __len__(self) -> int:
        """Return amount of key-value pairs."""
        return len(self._store)

    def __or__(self, other: Mapping) -> CaseInsensitiveDict:
        """Update operator for :py:class:`Mapping`'s.

        :return: A new :py:class`CaseInsensitiveDict` with the updated pairs of this and the other :py:class:`Mapping`.
        """
        if not isinstance(other, Mapping):
            return NotImplemented

        new = self.__class__(self)
        new.update(other)
        return new

    def __ror__(self, other: Mapping) -> CaseInsensitiveDict:
        """Update operator for :py:class:`Mapping`'s that don't support the syntax.

        :return: A new :py:class`CaseInsensitiveDict` with the updated pairs of this and the other :py:class:`Mapping`.
        """
        if not isinstance(other, Mapping):
            return NotImplemented

        new = self.__class__(other)
        new.update(self)
        return new

    def __ior__(self, other: Mapping) -> CaseInsensitiveDict:
        """Update setter operator for :py:class:`Mapping`'s.

        :return: This :py:class`CaseInsensitiveDict` with the updated pairs of the other :py:class:`Mapping`.
        """
        self.update(other)

        return self

    def lower_items(self) -> Generator[tuple[str, _VT], None, None]:
        """Like iteritems(), but with all lowercase keys."""
        return (
            (lower_key, keyval[1])
            for (lower_key, keyval)
            in self._store.items()
        )

    def __eq__(self, other: Mapping[str, _VT]) -> bool:
        """Compare items of other :py:class:`Mapping` case-insensitively."""
        if not isinstance(other, Mapping):
            return NotImplemented

        # Compare insensitively
        other_no_case: CaseInsensitiveDict[_VT] = self.__class__(other)
        return dict(self.lower_items()) == dict(other_no_case.lower_items())

    # Copy is required
    def copy(self) -> CaseInsensitiveDict[_VT]:
        """Return new :py:class:`CaseInsensitiveDict` with a copy of this instance's keys and values."""
        return self.__class__(self._store.values())

    def __repr__(self) -> str:
        """Representation of the :py:class:`CaseInsensitiveDict`."""
        return str(dict(self.items()))


class DeferredCallable(_AbstractCallable, Generic[_PT, _KT]):
    """A :py:class:`Callable` with args and kwargs stored for further execution.

    Supports deferred argument evaluation when using :py:class:`Callable`'s as arguments.
    This allows the value of the stored arguments to dynamically change depending on
    when the :py:class:`DeferredCallable` is executed.
    """

    __slots__ = ('_extra_pos_args', 'call_funcs', 'call_types', 'callable', 'args', 'kwargs')

    def __init__(self, __callable: Callable = lambda: None, /, *args: _PT | _PTCallable,
                 _extra_pos_args: int = 0,
                 _call_funcs: bool = True,
                 _call_types: bool = False,
                 **kwargs: _KT | _KTCallable) -> None:
        """Create a new :py:class:`DeferredCallable`.

        When called, any additional arguments must be expected with _extra_pos_args.
        Any arguments that exceed the extra positional argument limit will be trimmed off.

        :param __callable: Callable to store for later execution.
        :param args: positional arguments to store
        :param _extra_pos_args: Extra positional arguments to expect with self.run.
        :param _call_funcs: Whether to call (non-type) callable arguments
        :param _call_types: Whether to call class constructor arguments
        :param kwargs: keyword arguments to store
        """
        self._extra_pos_args: int = _extra_pos_args
        self.call_funcs: bool = _call_funcs
        self.call_types: bool = _call_types
        self.callable: Callable = __callable
        self.args: tuple[_PT | _PTCallable, ...] = args
        self.kwargs: dict[str, _KT | _KTCallable] = kwargs

    def __repr__(self) -> str:
        """Representation of the :py:class:`DeferredCallable` with the stored callable, args, and kwargs."""
        args, kwargs = self.args, self.kwargs
        return f'<{type(self).__name__} {self.callable} with {args=}, {kwargs=}>'

    def _evaluate_value(self, val: Any) -> Any:
        """Evaluate the given value to its called value if we want to know it.

        :param val: Value to evaluate.
        :return: The called value, if callable.
        """
        return val() if callable(val) and (
            (isinstance(val, type) and self.call_types) or
            (not isinstance(val, type) and self.call_funcs)
        ) else val

    def run(self, *args: Any, **kwargs: Any) -> Any:
        """Run the stored :py:class:`Callable`.

        Takes any additional arguments and temporarily adds to the stored arguments before execution.

        :param args: positional arguments to pass to callable.
        :param kwargs: keyword arguments to pass callable.
        :raises RuntimeError: Internal callable was not expecting the amount of positional arguments given.
        """
        # Add additional arguments from local args
        args = self.args + args[:self._extra_pos_args]  # Trim all arguments that are not expected
        kwargs |= self.kwargs

        # Evaluate all callable arguments
        args = tuple(self._evaluate_value(arg) for arg in args)
        kwargs = {key: self._evaluate_value(val) for key, val in kwargs.items()}

        try:
            return self.callable(*args, **kwargs)
        except TypeError as e:
            if ' positional argument but ' in str(e):
                raise RuntimeError(
                    f'{str(e).split(" ", maxsplit=1)[0]} was not expecting additional args, '
                    f'{type(self).__name__}._extra_call_args may not be set correctly.'
                ) from e


class DistributedCallable(_AbstractCallable, Generic[_CT, _PT, _KT]):
    """A :py:class:`Callable` that distributes arguments to all specified callables.

    Supports generic type hinting for the callable collections and arguments. Ex::

        foo: DistributedCallable[set] = DistributedCallable({bar}, 1, 2, 3, four=4)
        foo2: DistributedCallable[list] = DistributedCallable([bar], 1, 2, 3, four=4)

        # Now there's no error when doing

        foo.callables.add(baz)
        # And
        foo2.callables.append(baz)
    """

    __slots__ = ('callables', 'args', 'kwargs')

    def __init__(self, __callables: _CT = (), /, *args: _PT, **kwargs: _KT) -> None:
        """Create a new :py:class:`DistributedCallable` with the stored callables, args, and kwargs.

        _CT is a Type Generic containing a :py:class:`Collection` of callables.
        """
        self.callables: _CT = __callables
        self.args: tuple[_PT, ...] = args
        self.kwargs: dict[str, _KT] = kwargs

    def __repr__(self) -> str:
        """Representation of the :py:class:`DistributedCallable` with the stored callable, args, and kwargs."""
        args, kwargs = self.args, self.kwargs
        return f'<{type(self).__name__} {len(self.callables)} functions with {args=}, {kwargs=}>'

    def run(self, *args, **kwargs) -> tuple[Any, ...]:
        """Run all stored :py:class:`Callable`'s with the given extra arguments.

        :return: The results of each callable, packaged in a tuple.
        """
        # Add additional arguments from local args
        args = self.args + args
        kwargs = self.kwargs | kwargs

        results = tuple(func(*args, **kwargs) for func in self.callables)
        return results

    def generate(self, *args, **kwargs) -> Generator[Any, None, None]:
        """Run all stored :py:class:`Callable`'s with the given extra arguments. Yielding every result.

        :return: A generator yielding the results of each callable.
        """
        for func in self.callables:
            yield func(*args, **kwargs)


class Singleton:
    """A type which can only support one object instance at a time.

    All public "instances" of :py:class:`Singleton` are actually weakref proxies to
    the ``_Singleton__instance`` class attribute.

    DO:
        Anything with ``Singleton.instance()`` return value.

    DO NOT:
        Access ``_Singleton__instance`` class attribute outside the class.

        Store ``_Singleton__instance`` class attribute in any variable.

    -----

    When subclassing :py:class:`Singleton`, you must override the ``__init__`` abstract method.

    If you are subclassing multiple classes, it should always be the first in the order, and should specify the
    other class as your ``_singleton_base_type`` in your class definition. ex::

        class MySingleton(Singleton, OtherClass):
            _singleton_base_type = OtherClass
            _singleton_check_ref = True

            def __init__(self):
                ...

    If you are subclassing a C-Extension object (Such as a PySide ``QObject``) which keeps a hard-ref, or
    otherwise cannot guarantee the non-existence of outside references to the singleton instance, set the
    ``_singleton_check_ref`` attribute to ``False`` in your class definition.

    This removes all checks to the ``sys`` ref count.

    -----

    Create:
        To instantiate the :py:class:`Singleton` instance, use the ``create()`` class method.
        While instantiated,

        Attempting to use ``create()`` while the :py:class:`Singleton` is instantiated will
        raise a :py:class:`RuntimeError`.

    Access:
        You can access the :py:class:`Singleton` instance from the class definition by using the
        ``instance()`` class method.

        Attempting to use ``instance()`` before the :py:class:`Singleton` is created will
        raise a :py:class:`RuntimeError`.

        To check whether the :py:class:`Singleton` is instantiated, use the ``is_instantiated()`` class method.

    Destroy:
        :py:class:`Singleton` instances must be destroyed before you can create a new one. To do so,
        call the ``destroy()`` class method. Calling destroy guarantees that all references to the instance are
        deleted and garbage collected.

        If this isn't possible, and ``_singleton_check_ref`` is ``True``, a :py:class:`RuntimeError` is raised.

        Attempting to use ``destroy()`` before the :py:class:`Singleton` is created will
        raise a :py:class:`RuntimeError`.
    """

    __slots__: tuple[str, ...] = ()

    _singleton_base_type: type = object
    _singleton_check_ref: bool = True
    __instance: Singleton | None = None
    """The singular reference for the class instance. Should ONLY be accessed using weak reference proxies.

    Singleton.__instance is never used, all subclasses have their own __instance class attribute.
    """

    def __new__(cls, *args, **kwargs) -> NoReturn:
        """Call the :py:class:`Singleton`.create() classmethod instead.

        :raises TypeError: When called.
        """
        raise TypeError(
            f'Do not call default constructor for {cls.__name__}, instead call {cls.__name__}.create() explicitly.'
        )

    # pylint: disable=useless-super-delegation
    def __init__(self, *args, **kwargs) -> None:
        """Bare ``__init__`` used as placeholder.

        When replaced, its replacement is called by :py:class:`Singleton`.create()
        """
        super().__init__(*args, **kwargs)

    def __init_subclass__(cls, *args, **kwargs) -> None:
        """Ensure that the class instance is set to None for all subclasses."""
        super().__init_subclass__(*args, **kwargs)

        cls.__instance = None

    @classmethod
    def _check_ref_count(cls) -> bool:
        """Check if reference count is 1 for the :py:class:`Singleton` instance.

        Warns user if they access the internal instance.
        """
        if cls._singleton_check_ref:
            # 1 for cls.__instance, anymore is undefined behaviour. (Subtract 1 for argument reference)
            if cls.is_instantiated() and (sys.getrefcount(cls.__instance) - 1) > 1:
                warn(RuntimeWarning(
                    f'There is an outside reference to the internal {cls.__name__} instance. '
                    f'This is undefined behavior; please use weakref proxies.'
                ))
                return False

        return True

    @classmethod
    def create(cls, *args, **kwargs) -> None:
        """Preferred way of creating the :py:class:`Singleton` instance.

        :raises RuntimeError: If singleton instance currently exists.
        :raises TypeError: If __init__ is not overridden.
        """
        # All subclasses of Singleton should override __init__
        if cls.__init__ is Singleton.__init__:
            raise TypeError(
                f'Could not instantiate abstract class {cls.__name__} with abstract method __init__.'
            )

        if cls.is_instantiated():
            cls._check_ref_count()
            raise RuntimeError(
                f'Please destroy the {cls.__name__} singleton before creating a new {cls.__name__} instance.'
            )

        cls.__instance = cls._singleton_base_type.__new__(cls)  # type: ignore
        cls.__init__(cls.__instance, *args, **kwargs)

    @classmethod
    def instance(cls: type[_SingletonT]) -> _SingletonT:  # Real return type is weakref.ProxyType[Singleton]
        """Return a weak reference to the :py:class:`Singleton` instance.

        :raises RuntimeError: If singleton is not currently instantiated.
        """
        cls._check_ref_count()

        if not cls.is_instantiated():
            raise RuntimeError(
                f'Called {cls.__name__}.instance() when {cls.__name__} is not instantiated.'
            )

        return weakref.proxy(cls.__instance)

    @classmethod
    def destroy(cls) -> None:
        """Destroy the :py:class:`Singleton` instance.

        This results in the destruction of all weak references to the previous instance.

        :raises RuntimeError:
            If singleton instance could not be destroyed.
            If singleton is not currently instantiated.
        """
        if not cls._check_ref_count():
            raise RuntimeError(
                f'Could not destroy weak references to the {cls.__name__} instance. '
                f'Please remove all outside references to the internal instance before '
                f'calling {cls.__name__}.destroy().'
            )

        if not cls.is_instantiated():
            raise RuntimeError(
                f'Called {cls.__name__}.destroy() when {cls.__name__} is not instantiated.'
            )

        cls.__instance = None

    @classmethod
    def is_instantiated(cls) -> bool:
        """Return whether the :py:class:`Singleton` instance exists."""
        return isinstance(cls.__instance, cls)


_SingletonT = TypeVar('_SingletonT', bound=Singleton)
